from typing import Union
from biocutils import StringList, IntegerList, FloatList, BooleanList, NamedList
import numpy
import h5py
import os
import warnings

from .load_vector_from_hdf5 import load_vector_from_hdf5
from . import _utils_string as strings
from . import _utils_misc as misc


def read_atomic_vector(path: str, metadata: dict, atomic_vector_use_numeric_1darray: bool = False, **kwargs) -> Union[StringList, IntegerList, FloatList, BooleanList, numpy.ndarray]:
    """
    Read an atomic vector from its on-disk representation. In general, this
    function should not be called directly but instead via
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path: 
            Path to the directory containing the object.

        metadata: 
            Metadata for the object.

        atomic_vector_use_numeric_1darray: 
            Whether numeric vectors should be represented as 1-dimensional
            NumPy arrays. This is more memory-efficient than regular Python
            lists but discards the distinction between vectors and 1-D arrays.
            We set this to ``False`` by default to ensure that we can load
            names via :py:class:`~biocutils.NamedList.NamedList` subclasses.

        kwargs: 
            Further arguments, passed to nested objects.

    Returns:
        An atomic vector, represented as a
        :py:class:`~biocutils.StringList.StringList`,
        :py:class:`~biocutils.IntegerList.IntegerList`,
        :py:class:`~biocutils.FloatList.FloatList`,
        :py:class:`~biocutils.BooleanList.BooleanList`.
    """

    with h5py.File(os.path.join(path, "contents.h5"), "r") as handle:
        ghandle = handle["atomic_vector"]
        vectype = strings.load_scalar_string_attribute_from_hdf5(ghandle, "type")
        dhandle = ghandle["values"]

        expected_type = misc.translate_type(vectype)
        output = load_vector_from_hdf5(dhandle, expected_type, atomic_vector_use_numeric_1darray)

        if "names" in ghandle:
            if isinstance(output, NamedList):
                output.set_names(strings.load_string_vector_from_hdf5(ghandle["names"]), in_place=True)
            else:
                warnings.warn("skipping names when reading atomic vectors as 1-dimensional NumPy arrays")
        return output



from typing import List
import h5py


def save_fixed_length_strings(handle: h5py.Group, name: str, x: List[str]) -> h5py.Dataset:
    """Save a list of strings into a fixed-length string dataset.

    Args:
        handle:     
            Handle to a HDF5 Group.
        
        name: 
            Name of the dataset to create in ``handle``.
        
        x: 
            List of strings to save.

    Returns:
        ``x`` is saved into the group as a fixed-length string dataset,
        and a NumPy dataset handle is returned.
    """
    tmp = [ y.encode("UTF8") for y in x ]
    maxed = 1
    for b in tmp:
        if len(b) > maxed:
            maxed = len(b)
    return handle.create_dataset(name, data=tmp, dtype="S" + str(maxed), compression="gzip", chunks=True)


def load_string_vector_from_hdf5(handle: h5py.Dataset) -> List[str]:
    output = list(handle)
    if len(output) and isinstance(output[0], bytes):
        for i, x in enumerate(output):
            output[i] = x.decode("UTF-8")
    return output


def load_scalar_string_attribute_from_hdf5(handle, name: str) -> str:
    output = handle.attrs[name]
    if isinstance(output, bytes):
        output = output.decode("UTF-8")
    return output



from typing import Any, Union, Literal
import numpy as np
from warnings import warn
from functools import singledispatch
from biocutils import Factor, StringList, NamedList, IntegerList, BooleanList, FloatList
import os
import json
import gzip
import h5py

from .save_object import save_object, validate_saves
from .alt_save_object import alt_save_object
from . import _utils_misc as misc
from . import _utils_string as strings
from . import write_vector_to_hdf5 as write


@save_object.register
@validate_saves
def save_simple_list_from_dict(x: dict, path: str, simple_list_mode: Literal["hdf5", "json"] = "json", **kwargs):
    """Method for saving dictionaries (Python analogues to R-style named lists)
    to the corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to a directory in which to save the object.

        simple_list_mode: 
            Whether to save in HDF5 or JSON mode.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    _save_simple_list_internal(x, path, simple_list_mode, **kwargs)
    return


@save_object.register
@validate_saves
def save_simple_list_from_list(x: list, path: str, simple_list_mode: Literal["hdf5", "json"] = "json", **kwargs):
    """Method for saving lists (Python analogues to R-style unnamed lists) to
    the corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to a directory in which to save the object.

        simple_list_mode: 
            Whether to save in HDF5 or JSON mode.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    _save_simple_list_internal(x, path, simple_list_mode, **kwargs)
    return


@save_object.register
@validate_saves
def save_simple_list_from_NamedList(x: NamedList, path: str, simple_list_mode: Literal["hdf5", "json"] = "json", **kwargs):
    """Method for saving a NamedList to its corresponding file representation,
    see :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to a directory in which to save the object.

        simple_list_mode: 
            Whether to save in HDF5 or JSON mode.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    _save_simple_list_internal(x, path, simple_list_mode, **kwargs)
    return


##########################################################################


def _save_simple_list_internal(x: Union[dict, list, NamedList], path: str, simple_list_mode: Literal["hdf5", "json"] = None, **kwargs):
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        format2 = simple_list_mode 
        if format2 == "json":
            format2 = "json.gz"
        handle.write('{ "type": "simple_list", "simple_list": { "version": "1.0", "format": "' + format2 + '" } }')

    externals = []

    if simple_list_mode == "json":
        transformed = _save_simple_list_recursive(x, externals, None)
        transformed["version"] = "1.2"
        opath = os.path.join(path, "list_contents.json.gz")
        with gzip.open(opath, "wt") as handle:
            json.dump(transformed, handle)

    else:
        opath = os.path.join(path, "list_contents.h5")
        with h5py.File(opath, "w") as handle:
            ghandle = handle.create_group("simple_list")
            ghandle.attrs["uzuki_version"] = "1.3"
            _save_simple_list_recursive(x, externals, ghandle)

    if len(externals):
        exdir = os.path.join(path, "other_contents")
        os.mkdir(exdir)
        for i, ex in enumerate(externals):
            alt_save_object(ex, os.path.join(exdir, str(i)), **kwargs)
    return


@singledispatch
def _save_simple_list_recursive(x: Any, externals: list, handle):
    return _save_simple_list_recursive_Any(x, externals, handle)


def _save_simple_list_recursive_Any(x: Any, externals: list, handle):
    externals.append(x)
    if handle is None:
        return { "type": "external", "index": len(externals) - 1 }
    else:
        handle.attrs["uzuki_object"] = "external"
        handle.create_dataset("index", data=len(externals) - 1, dtype='i4')
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_StringList(x: StringList, externals: list, handle):
    nms = x.get_names()

    if handle is None:
        output = { "type": "string", "values": x.as_list() }
        if nms is not None:
            output["names"] = nms.as_list()
        return output

    handle.attrs["uzuki_object"] = "vector"
    handle.attrs["uzuki_type"] = "string"
    write.write_string_vector_to_hdf5(handle, "data", x.as_list())
    if nms is not None:
        strings.save_fixed_length_strings(handle, "names", nms.as_list())
    return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_IntegerList(x: IntegerList, externals: list, handle):
    nms = x.get_names()

    if handle is None:
        final_type = "integer"
        if misc.sequence_exceeds_int32(x):
            final_type = "number"
        output = { "type": final_type, "values": x.as_list() }
        if nms is not None:
            output["names"] = nms.as_list()
        return output

    handle.attrs["uzuki_object"] = "vector"
    dset = write.write_integer_vector_to_hdf5(handle, "data", x.as_list(), allow_float_promotion=True)
    if np.issubdtype(dset, np.floating):
        handle.attrs["uzuki_type"] = "number"
    else:
        handle.attrs["uzuki_type"] = "integer"
    if nms is not None:
        strings.save_fixed_length_strings(handle, "names", nms.as_list())
    return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_FloatList(x: FloatList, externals: list, handle):
    nms = x.get_names()

    if handle is None:
        xcopy = [ _sanitize_masked_float_json(y) for y in x.as_list() ]
        output = { "type": "number", "values": xcopy }
        if nms is not None:
            output["names"] = nms.as_list()
        return output

    handle.attrs["uzuki_object"] = "vector"
    handle.attrs["uzuki_type"] = "number"
    write.write_float_vector_to_hdf5(handle, "data", x.as_list())
    if nms is not None:
        strings.save_fixed_length_strings(handle, "names", nms.as_list())
    return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_BooleanList(x: BooleanList, externals: list, handle):
    nms = x.get_names()

    if handle is None:
        output = { "type": "boolean", "values": x.as_list() }
        if nms is not None:
            output["names"] = nms.as_list()
        return output

    handle.attrs["uzuki_object"] = "vector"
    handle.attrs["uzuki_type"] = "boolean"
    write.write_boolean_vector_to_hdf5(handle, "data", x.as_list())
    if nms is not None:
        strings.save_fixed_length_strings(handle, "names", nms.as_list())
    return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_list(x: list, externals: list, handle):
    if handle is None:
        vals = []
        collected = { "type": "list", "values": vals }
        for i, y in enumerate(x):
            vals.append(_save_simple_list_recursive(y, externals, None))
        return collected
    else:
        handle.attrs["uzuki_object"] = "list"
        dhandle = handle.create_group("data")
        for i, y in enumerate(x):
            ghandle = dhandle.create_group(str(i))
            _save_simple_list_recursive(y, externals, ghandle)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_dict(x: dict, externals: list, handle):
    if handle is None:
        vals = []
        names = []
        collected = { "type": "list", "values": vals, "names": names }
        for k, v in x.items():
            if not isinstance(k, str):
                warn("converting non-string key with value " + str(k) + " to a string", UserWarning)
            names.append(str(k))
            vals.append(_save_simple_list_recursive(v, externals, None))
        return collected
    else:
        handle.attrs["uzuki_object"] = "list"
        dhandle = handle.create_group("data")
        names = []
        for k, v in x.items():
            ghandle = dhandle.create_group(str(len(names)))
            _save_simple_list_recursive(v, externals, ghandle)
            if not isinstance(k, str):
                warn("converting non-string key with value " + str(k) + " to a string", UserWarning)
            names.append(str(k))
        strings.save_fixed_length_strings(handle, "names", names)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_NamedList(x: NamedList, externals: list, handle):
    if x.get_names() is None:
        return _save_simple_list_recursive_list(x.as_list(), externals, handle)

    if handle is None:
        vals = []
        collected = { "type": "list", "values": vals, "names": x.get_names().as_list() }
        for v in x.as_list():
            vals.append(_save_simple_list_recursive(v, externals, None))
        return collected
    else:
        handle.attrs["uzuki_object"] = "list"
        dhandle = handle.create_group("data")
        for i, v in enumerate(x.as_list()):
            ghandle = dhandle.create_group(str(i))
            _save_simple_list_recursive(v, externals, ghandle)
        strings.save_fixed_length_strings(handle, "names", x.get_names().as_list())
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_bool(x: bool, externals: list, handle):
    if handle is None:
        return { "type": "boolean", "values": bool(x) }
    else:
        _save_scalar_hdf5(handle, x=x, dtype=bool)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_int(x: int, externals: list, handle):
    if not misc.scalar_exceeds_int32(x):
        if handle is None:
            return { "type": "integer", "values": int(x) }
        else:
            _save_scalar_hdf5(handle, x=x, dtype=int)
            return
    else:
        if handle is None:
            return { "type": "number", "values": x }
        else:
            _save_scalar_hdf5(handle, x=x, dtype=float)
            return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_str(x: str, externals: list, handle):
    if handle is None:
        return { "type": "string", "values": str(x) }
    else:
        _save_scalar_hdf5(handle, x=x, dtype=str)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_float(x: float, externals: list, handle):
    if handle is None:
        return { "type": "number", "values": _sanitize_float_json(x) }
    else:
        _save_scalar_hdf5(handle, x=x, dtype=float)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_MaskedConstant(x: np.ma.core.MaskedConstant, externals: list, handle):
    if handle is None:
        return { "type": "number", "values": None}
    else:
        _save_scalar_hdf5(handle, x=np.NaN, dtype=float, missing_placeholder=np.NaN)
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_MaskedConstant(x: np.ndarray, externals: list, handle):
    if len(x.shape) == 0:
        return _save_simple_list_recursive(x[()], externals, handle)
    else:
        return _save_simple_list_recursive_Any(x, externals, handle)


@_save_simple_list_recursive.register
def _save_simple_list_recursive_numpy_generic(x: np.generic, externals: list, handle):
    final_type = None
    if np.issubdtype(x.dtype, np.integer):
        if not misc.scalar_exceeds_int32(x):
            final_type = int
        else:
            final_type = float
    elif np.issubdtype(x.dtype, np.floating):
        final_type = float
    elif x.dtype == np.bool_:
        final_type = bool

    if final_type == int:
        if handle is None:
            return { "type": "integer", "values": int(x) }
        else:
            _save_scalar_hdf5(handle, x=x, dtype=int)
            return
    elif final_type == float:
        if handle is None:
            return { "type": "number", "values": _sanitize_float_json(x) }
        else:
            _save_scalar_hdf5(handle, x=x, dtype=float)
            return
    elif final_type == bool:
        if handle is None:
            return { "type": "boolean", "values": bool(x) }
        else:
            _save_scalar_hdf5(handle, x=x, dtype=bool)
            return 
    else:
        raise NotImplementedError("no staging method for NumPy array scalars of " + str(x.dtype))


@_save_simple_list_recursive.register
def _save_simple_list_recursive_factor(x: Factor, externals: list, handle):
    nms = x.get_names()

    if handle is None:
        output = { 
            "type": "factor",
            "values": [(None if y == -1 else int(y)) for y in x.get_codes()],
            "levels": x.get_levels().as_list(),
            "ordered": x.get_ordered(),
        }
        if nms is not None:
            output["names"] = nms.as_list()
        return output

    else:
        handle.attrs["uzuki_object"] = "vector"
        handle.attrs["uzuki_type"] = "factor"

        dhandle = handle.create_dataset("data", data=x.get_codes(), dtype="i4", compression="gzip", chunks=True)
        if (x.get_codes() == -1).any():
            dhandle.attrs.create("missing-value-placeholder", data=-1, dtype="i4")

        strings.save_fixed_length_strings(handle, "levels", x.get_levels().as_list())
        if x.get_ordered():
            handle.create_dataset("ordered", data=x.get_ordered(), dtype="i1")

        if nms is not None:
            strings.save_fixed_length_strings(handle, "names", nms.as_list())
        return


@_save_simple_list_recursive.register
def _save_simple_list_recursive_none(x: None, externals: list, handle):
    if handle is None:
        return { "type": "nothing" }
    else:
        handle.attrs["uzuki_object"] = "nothing"
        return


##########################################################################


def _sanitize_float_json(x):
    if np.isnan(x):
        return "NaN"
    elif x == np.Inf:
        return "Inf"
    elif x == -np.Inf:
        return "-Inf"
    return float(x)


def _sanitize_masked_float_json(x):
    if x is None:
        return None
    return _sanitize_float_json(x)


##########################################################################


def _save_scalar_hdf5(handle, x, dtype, missing_placeholder = None):
    handle.attrs["uzuki_object"] = "vector"

    if dtype == bool:
        handle.attrs["uzuki_type"] = "boolean"
        savetype = 'i1'
    elif dtype == int:
        handle.attrs["uzuki_type"] = "integer"
        savetype = 'i4'
    elif dtype == str:
        handle.attrs["uzuki_type"] = "string"
        savetype = None
    elif dtype == float:
        handle.attrs["uzuki_type"] = "number"
        savetype = 'f8'
    else:
        raise NotImplementedError("no staging method for scalars of " + str(dtype))

    dhandle = handle.create_dataset("data", data=x, dtype=savetype)
    if missing_placeholder:
        dhandle.attrs.create("missing-value-placeholder", data=missing_placeholder, dtype=savetype)



import dolomite_base as dl
import numpy


def test_choose_missing_integer_placeholder():
    out = dl.choose_missing_integer_placeholder([None, 1, 2, 3]) 
    assert out.dtype == numpy.int32
    assert out == -2**31

    out = dl.choose_missing_integer_placeholder([None, 1, 2, -2**31]) 
    assert out.dtype == numpy.int32
    assert out == 2**31 - 1

    out = dl.choose_missing_integer_placeholder([None, 1, 2, 2**31-1, -2**31]) 
    assert out.dtype == numpy.int32
    assert out == 0

    out = dl.choose_missing_integer_placeholder([None, 1, 2, 2**31-1, 0, -2**31]) 
    assert out.dtype == numpy.int32
    assert out == -2**31 + 1

    # Attempts to respect the provided NumPy type.
    out = dl.choose_missing_integer_placeholder(numpy.array([1, 2], dtype=numpy.uint8))
    assert out.dtype == numpy.uint8
    assert out == 0

    out = dl.choose_missing_integer_placeholder(numpy.array([1, 2, 0], dtype=numpy.uint8))
    assert out.dtype == numpy.uint8
    assert out == 255

    out = dl.choose_missing_integer_placeholder(numpy.array(range(256), dtype=numpy.uint8))
    assert out.dtype == numpy.int32
    assert out == -2**31

    out = dl.choose_missing_integer_placeholder(numpy.array(range(256), dtype=numpy.uint8), max_dtype=numpy.uint8)
    assert out is None

    # Works with a set.
    out = dl.choose_missing_integer_placeholder(set([1, 2, 2**31-1, 0, -2**31, -2**31 + 1]))
    assert out.dtype == numpy.int32
    assert out == -2**31 + 2

    # Works with a Numpy masked arrays.
    out = dl.choose_missing_integer_placeholder(numpy.ma.MaskedArray([1, 2, 2**31-1, 0, -2**31], mask=[True] + [False] * 4))
    assert out.dtype == numpy.int32
    assert out == -2**31 + 1 


def test_choose_missing_float_placeholder():
    out = dl.choose_missing_float_placeholder([None, 1, 2, 3]) 
    assert out.dtype == numpy.float64
    assert numpy.isnan(out)

    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan]) 
    assert out.dtype == numpy.float64
    assert out == numpy.inf

    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan, numpy.inf]) 
    assert out.dtype == numpy.float64
    assert out == -numpy.inf

    fstats = numpy.finfo(numpy.float64)
    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan, numpy.inf, -numpy.inf]) 
    assert out.dtype == numpy.float64
    assert out == fstats.min

    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min]) 
    assert out.dtype == numpy.float64
    assert out == fstats.max

    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min, fstats.max]) 
    assert out.dtype == numpy.float64
    assert out == 0

    out = dl.choose_missing_float_placeholder([None, 1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min, fstats.max, 0]) 
    assert out.dtype == numpy.float64
    assert out == fstats.min / 2

    # Respects original type. 
    out = dl.choose_missing_float_placeholder(numpy.array([1, 2, 3], dtype=numpy.float32)) 
    assert out.dtype == numpy.float32
    assert numpy.isnan(out)

    fstats = numpy.finfo(numpy.float32)
    out = dl.choose_missing_float_placeholder(numpy.array([1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min, fstats.max, 0], dtype=numpy.float32)) 
    assert out.dtype == numpy.float32
    assert out == fstats.min / 2

    # Works with NumPy masked arrays.
    out = dl.choose_missing_float_placeholder(numpy.ma.MaskedArray(numpy.array([1, 2, 3, numpy.nan]), mask=[True] + [False] * 3))
    assert out.dtype == numpy.float64
    assert out == numpy.inf

    fstats = numpy.finfo(numpy.float64)
    out = dl.choose_missing_float_placeholder(numpy.ma.MaskedArray(numpy.array([1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min, fstats.max, 0]), mask=[True] + [False] * 8))
    assert out.dtype == numpy.float64
    assert out == fstats.min/2

    # Works with a set.
    out = dl.choose_missing_float_placeholder(set([None, 1, 2, 3]))
    assert out.dtype == numpy.float64
    assert numpy.isnan(out)

    fstats = numpy.finfo(numpy.float64)
    out = dl.choose_missing_float_placeholder(set([None, 1, 2, 3, numpy.nan, numpy.inf, -numpy.inf, fstats.min, fstats.max, 0]))
    assert out.dtype == numpy.float64
    assert fstats.min / 2


def test_choose_missing_string_placeholder():
    out = dl.choose_missing_string_placeholder([None, "XA"])
    assert out == "NA"

    out = dl.choose_missing_string_placeholder([None, "NA"])
    assert out == "NA_"

    out = dl.choose_missing_string_placeholder(set(["NA", "NA_"]))
    assert out == "NA__"

    out = dl.choose_missing_string_placeholder(numpy.array(["NA", "NA_"]))
    assert out == "NA__"



import dolomite_base as dl
import numpy as np
from tempfile import mkdtemp
from biocframe import BiocFrame
from biocutils import Factor, StringList, NamedList, FloatList, IntegerList, BooleanList
import os


def test_simple_list_basic():
    everything = {
        "i_am_a_string": "foo",
        "i_am_a_number": 1.23,
        "i_am_a_integer": -23,
        "i_am_a_boolean": False,
        "i_am_a_list": [ 1, True, 2.3, "bar" ],
        "i_am_a_dict": {
            "string": StringList([ "b", "c", "d", "e" ]),
            "float": FloatList([ 1.5, 2.5, 3.5, 4.5, 5.5 ]),
            "int": IntegerList([ 1, 2, 3, 4, 5 ]),
            "bool": BooleanList([ True, False, False, True, False ])
        },
        "i_am_nothing": None
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    dl.save_object(everything, dir, simple_list_mode="json")
    assert os.path.exists(os.path.join(dir, "list_contents.json.gz"))

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert everything["i_am_a_string"] == roundtrip["i_am_a_string"]
    assert everything["i_am_a_number"] == roundtrip["i_am_a_number"]
    assert everything["i_am_a_integer"] == roundtrip["i_am_a_integer"]
    assert everything["i_am_a_boolean"] == roundtrip["i_am_a_boolean"]
    assert everything["i_am_a_list"] == roundtrip["i_am_a_list"].as_list()
    assert everything["i_am_nothing"] == roundtrip["i_am_nothing"]

    assert everything["i_am_a_dict"]["string"] == roundtrip["i_am_a_dict"]["string"]
    assert everything["i_am_a_dict"]["float"] == roundtrip["i_am_a_dict"]["float"]
    assert everything["i_am_a_dict"]["int"] == roundtrip["i_am_a_dict"]["int"]
    assert everything["i_am_a_dict"]["bool"] == roundtrip["i_am_a_dict"]["bool"]

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(everything, dir, simple_list_mode="hdf5")
    assert os.path.exists(os.path.join(dir, "list_contents.h5"))

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert everything["i_am_a_string"] == roundtrip["i_am_a_string"]
    assert everything["i_am_a_number"] == roundtrip["i_am_a_number"]
    assert everything["i_am_a_integer"] == roundtrip["i_am_a_integer"]
    assert everything["i_am_a_boolean"] == roundtrip["i_am_a_boolean"]
    assert everything["i_am_a_list"] == roundtrip["i_am_a_list"].as_list()
    assert everything["i_am_nothing"] == roundtrip["i_am_nothing"]

    assert everything["i_am_a_dict"]["string"] == roundtrip["i_am_a_dict"]["string"]
    assert everything["i_am_a_dict"]["float"] == roundtrip["i_am_a_dict"]["float"]
    assert everything["i_am_a_dict"]["int"] == roundtrip["i_am_a_dict"]["int"]
    assert everything["i_am_a_dict"]["bool"] == roundtrip["i_am_a_dict"]["bool"]


def test_simple_list_unnamed():
    everything = [ "foo", 1, 2, False, None ]

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert roundtrip.get_names() is None
    assert roundtrip.as_list() == everything

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5 ")
    dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert roundtrip.get_names() is None
    assert roundtrip.as_list() == everything


def test_simple_list_NamedList():
    everything = NamedList([NamedList([ "foo", 1, 2, False, None ]), "FOO"], names=["A", ""])

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert roundtrip.get_names().as_list() == ["A", ""]
    assert roundtrip[""] == "FOO"
    assert isinstance(roundtrip["A"], NamedList)
    assert roundtrip["A"] == everything["A"]

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5 ")
    dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, NamedList)
    assert roundtrip.get_names().as_list() == ["A", ""]
    assert roundtrip[""] == "FOO"
    assert isinstance(roundtrip["A"], NamedList)
    assert roundtrip["A"] == everything["A"]


def test_simple_list_numpy_scalars():
    everything = {
        "float": np.float64(9.9),
        "int": np.int8(10),
        "bool": np.bool_(False),
        "float2": np.array(1.45),
        "int2": np.array(-5),
        "bool2": np.array(True),
        "masked": np.ma.masked
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip["float"], float)
    assert roundtrip["float"] == 9.9
    assert isinstance(roundtrip["int"], int)
    assert roundtrip["int"] == 10
    assert isinstance(roundtrip["bool"], bool)
    assert isinstance(roundtrip["float2"], float)
    assert roundtrip["float2"] == 1.45
    assert isinstance(roundtrip["int2"], int)
    assert roundtrip["int2"] == -5
    assert isinstance(roundtrip["bool2"], bool)
    assert roundtrip["bool2"]
    assert roundtrip["masked"] is None

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip["float"], float)
    assert roundtrip["float"] == 9.9
    assert isinstance(roundtrip["int"], int)
    assert roundtrip["int"] == 10
    assert isinstance(roundtrip["bool"], bool)
    assert roundtrip["bool"] == False
    assert isinstance(roundtrip["float2"], float)
    assert roundtrip["float2"] == 1.45
    assert isinstance(roundtrip["int2"], int)
    assert roundtrip["int2"] == -5
    assert isinstance(roundtrip["bool2"], bool)
    assert roundtrip["bool2"]
    assert roundtrip["masked"] is None


def test_simple_list_large_integers():
    everything = {
        "a": 2**31 - 1,
        "b": 2**31,
        "c": IntegerList([-2**32])
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    meta = dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert everything["a"] == roundtrip["a"]
    assert isinstance(roundtrip["a"], int)
    assert everything["b"] == roundtrip["b"]
    assert isinstance(roundtrip["b"], float)
    assert everything["c"].as_list() == roundtrip["c"].as_list()
    assert isinstance(roundtrip["c"], FloatList)

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    meta = dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert everything["a"] == roundtrip["a"]
    assert isinstance(roundtrip["a"], int)
    assert everything["b"] == roundtrip["b"]
    assert isinstance(roundtrip["b"], float)
    assert everything["c"].as_list() == roundtrip["c"].as_list()
    assert isinstance(roundtrip["c"], FloatList)


def test_simple_list_special_float():
    everything = {
        "a": np.NaN,
        "b": FloatList([np.Inf, -np.Inf, np.NaN])
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    meta = dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert np.isnan(roundtrip["a"])
    assert roundtrip["b"][0] == np.Inf
    assert roundtrip["b"][1] == -np.Inf
    assert np.isnan(roundtrip["b"][2])

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    meta = dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert np.isnan(roundtrip["a"])
    assert roundtrip["b"][0] == np.Inf
    assert roundtrip["b"][1] == -np.Inf
    assert np.isnan(roundtrip["b"][2])


def test_simple_list_external():
    everything = {
        "a": BiocFrame({ "a_1": [ 1, 2, 3 ], "a_2": [ "A", "B", "C" ] }),
        "b": BiocFrame(number_of_rows=10),
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    dl.save_object(everything, dir, simple_list_mode='json')
    roundtrip = dl.read_object(dir)
    assert roundtrip["a"].get_column_names().as_list() == [ "a_1", "a_2" ]
    assert roundtrip["a"].shape[0] == 3
    assert roundtrip["b"].get_column_names().as_list() == []
    assert roundtrip["b"].shape[0] == 10

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(everything, dir, simple_list_mode="hdf5")
    roundtrip = dl.read_object(dir)
    assert roundtrip["a"].get_column_names().as_list() == [ "a_1", "a_2" ]
    assert roundtrip["a"].shape[0] == 3
    assert roundtrip["b"].get_column_names().as_list() == []
    assert roundtrip["b"].shape[0] == 10


def test_simple_list_factor():
    everything = {
        "regular": Factor.from_sequence([ "sydney", "brisbane", "sydney", "melbourne"]),
        "missing": Factor.from_sequence([ "sydney", None, "sydney", None]),
        "ordered": Factor.from_sequence([ "sydney", "brisbane", "sydney", "melbourne"], levels=["sydney", "melbourne", "brisbane"], ordered=True),
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    meta = dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip["regular"], Factor)
    assert list(roundtrip["regular"]) == list(everything["regular"])
    assert not roundtrip["regular"].get_ordered()
    assert list(roundtrip["missing"]) == list(everything["missing"])
    assert list(roundtrip["ordered"]) == list(everything["ordered"])
    assert roundtrip["ordered"].get_ordered()

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    meta = dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip["regular"], Factor)
    assert list(roundtrip["regular"]) == list(everything["regular"])
    assert not roundtrip["regular"].get_ordered()
    assert list(roundtrip["missing"]) == list(everything["missing"])
    assert list(roundtrip["ordered"]) == list(everything["ordered"])
    assert roundtrip["ordered"].get_ordered()


def test_simple_list_named():
    everything = {
        "factor": Factor.from_sequence([ "sydney", "brisbane", "sydney", "melbourne"], names=["A", "B", "C", "D"]),
        "string": StringList(["Aria", "Akari", "Akira", "Aika"], names=["1", "2", "3", "4"])
    }

    # Stage as JSON.
    dir = os.path.join(mkdtemp(), "json")
    meta = dl.save_object(everything, dir, simple_list_mode="json")

    roundtrip = dl.read_object(dir)
    assert list(roundtrip["factor"]) == list(everything["factor"])
    assert roundtrip["factor"].get_names() == everything["factor"].get_names()
    assert list(roundtrip["string"]) == list(everything["string"])
    assert roundtrip["string"].get_names() == everything["string"].get_names()

    # Stage as HDF5.
    dir = os.path.join(mkdtemp(), "hdf5")
    meta = dl.save_object(everything, dir, simple_list_mode="hdf5")

    roundtrip = dl.read_object(dir)
    assert list(roundtrip["factor"]) == list(everything["factor"])
    assert roundtrip["factor"].get_names() == everything["factor"].get_names()
    assert list(roundtrip["string"]) == list(everything["string"])
    assert roundtrip["string"].get_names() == everything["string"].get_names()



from typing import Tuple, Sequence, Optional
import numpy


def _scan_for_integer_placeholder(x: Sequence[int], dtype: type) -> Optional[numpy.generic]:
    stats = numpy.iinfo(dtype)
    if not stats.min in x:
        return dtype(stats.min)
    if not stats.max in x:
        return dtype(stats.max)
    if stats.min != 0 and not 0 in x:
        return dtype(0)

    if not isinstance(x, set):
        alt = set()
        for y in x:
            if y is not None and not numpy.ma.is_masked(y):
                alt.add(y)
        x = alt

    for i in range(stats.min + 1, stats.max):
        if not i in x:
            return dtype(i)
    return None


def choose_missing_integer_placeholder(x: Sequence[int], max_dtype: type = numpy.int32) -> Optional[numpy.generic]:
    """Choose a missing placeholder for integer sequences.

    Args:
        x:
            Sequence of integers, possibly containing masked or None values.

        max_dtype: 
            Integer NumPy type that is guaranteed to faithfully represent all
            (non-None, non-masked) values of ``x``.

    Returns:
        Value of the placeholder. This is guaranteed to be of a type that can
        fit into ``max_dtype``. It also may not be of the same type as
        ``x.dtype`` if ``x`` is a NumPy array, so some casting may be required
        when replacing missing values with the placeholder.

        If no suitable placeholder can be found, None is returned instead.
    """
    if isinstance(x, numpy.ndarray) and (x.itemsize < max_dtype().itemsize):
        candidate = _scan_for_integer_placeholder(x, x.dtype.type)
        if not candidate is None:
            return candidate
    return _scan_for_integer_placeholder(x, max_dtype)


def choose_missing_float_placeholder(x: Sequence[float], dtype: type = numpy.float64) -> Optional[numpy.generic]:
    """Choose a missing placeholder for float sequences.

    Args:
        x: 
            Sequence of floats, possibly containing masked or None values.

        dtype:
            Floating-point NumPy type to use for the placeholder. Ignored if
            ``x`` is already a NumPy floating-point array, in which case the
            ``dtype`` is just set to the ``x.dtype``.

    Returns:
        Value of the placeholder. If ``x`` is a NumPy floating-point array,
        this is guaranteed to be of the same type as ``x.dtype``.

        If no suitable placeholder can be found, None is returned instead.
    """
    if isinstance(x, numpy.ndarray) and numpy.issubdtype(x.dtype, numpy.floating):
        dtype = x.dtype.type

    can_nan = True
    for y in x:
        if y is not None and not numpy.ma.is_masked(y) and numpy.isnan(y):
            can_nan = False
            break
    if can_nan:
        return dtype(numpy.NaN)

    if not numpy.inf in x:
        return dtype(numpy.inf)
    if not -numpy.inf in x:
        return dtype(-numpy.inf)

    stats = numpy.finfo(dtype)
    if not stats.min in x:
        return dtype(stats.min)
    if not stats.max in x:
        return dtype(stats.max)
    if not 0 in x:
        return dtype(0)

    if not isinstance(x, set):
        alt = set()
        for y in x:
            if y is not None and not numpy.ma.is_masked(y):
                alt.add(y)
        x = alt

    accumulated = []
    for y in x:
        if y is not None and numpy.isfinite(y):
            accumulated.append(y)
    accumulated.sort()
    for i in range(1, len(accumulated)):
        previous = accumulated[i-1]
        current = accumulated[i]
        mid = previous + (current - previous) / dtype(2)
        if mid != previous and mid != current:
            return mid

    # Highly unlikely that we'll get to this point.
    return None


def choose_missing_string_placeholder(x: Sequence[str]) -> str:
    """Choose a missing placeholder for string sequences.

    Args:
        x: 
            Sequence of strings, possibly containing missing or None values.

    Returns:
        String to use as the placeholder. This may be longer than the maximum
        string length in ``x`` (for fixed-length-string arrays), so some
        casting may be required.
    """
    placeholder = "NA"
    while placeholder in x:
        placeholder += "_"
    return placeholder



from typing import Union
import os

from .alt_read_object import alt_read_object
from . import lib_dolomite_base as lib


def read_simple_list(path: str, metadata: dict, **kwargs) -> Union[dict, list]:
    """Read an R-style list from its on-disk representation in the **uzuki2**
    format.  In general, this function should not be called directly but
    instead via :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path: 
            Path to the directory containing the object.

        metadata: 
            Metadata for the object.
    
        kwargs: 
            Further arguments, passed to nested objects.

    Returns:
        A list or dictionary.
    """

    other_dir = os.path.join(path, "other_contents")
    children = []
    if os.path.exists(other_dir):
        files = os.listdir(other_dir)
        collected = []
        for f in files:
            if f.isdigit():
                collected.append(f)
        children = [None] * len(collected)
        for f in collected:
            children[int(f)] = alt_read_object(os.path.join(other_dir, f))

    if metadata["simple_list"]["format"] == "hdf5":
        full_path = os.path.join(path, "list_contents.h5")
        return lib.load_list_hdf5(full_path, "simple_list", children)
    else:
        full_path = os.path.join(path, "list_contents.json.gz")
        return lib.load_list_json(full_path, children)



from biocframe import BiocFrame
from biocutils import Factor, StringList, IntegerList, BooleanList, FloatList
import dolomite_base as dl
import numpy as np
import os
import h5py
from tempfile import mkdtemp


def test_data_frame_list():
    df = BiocFrame({
        "akari": [ 1, 2, 3, 4, 5 ],
        "aika": [ "sydney", "melbourne", "", "perth", "adelaide" ],
        "alice": [ True, False, False, True, True ],
        "ai": [ 2.3, 1.2, 5.2, 3.1, -1.2 ],
        "alicia": [ 5, 2, 1, 3.2, -2 ], # mixed integers and numbers
        "akira": StringList(["A", "B", "C", "D", "E"]),
    })

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BiocFrame)

    assert list(roundtrip.get_column("akari")) == df.get_column("akari")
    assert roundtrip.get_column("akari").dtype.type == np.int32

    assert isinstance(roundtrip.get_column("aika"), StringList)
    assert roundtrip.get_column("aika").as_list() == df.get_column("aika")

    assert list(roundtrip.get_column("alice")) == df.get_column("alice")
    assert roundtrip.get_column("alice").dtype.type == np.bool_

    assert list(roundtrip.get_column("ai")) == df.get_column("ai")
    assert roundtrip.get_column("ai").dtype.type == np.float64

    assert list(roundtrip.get_column("alicia")) == df.get_column("alicia")
    assert roundtrip.get_column("alicia").dtype.type == np.float64

    assert roundtrip.get_column("akira") == df.get_column("akira")
    assert isinstance(roundtrip.get_column("akira"), StringList)


def test_data_frame_external_list():
    df = BiocFrame({
        "akari": [ 1, 2, 3, 4, 5 ],
        "aika": [ "sydney", "melbourne", "", "perth", "adelaide" ],
        "alice": [ True, False, False, True, True ],
        "ai": [ 2.3, 1.2, 5.2, 3.1, -1.2 ],
    })

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir, data_frame_convert_list_to_vector=False)
    assert os.path.exists(os.path.join(dir, "other_columns", "0"))

    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BiocFrame)
    assert roundtrip.get_column("akari").as_list() == df.get_column("akari")
    assert roundtrip.get_column("aika").as_list() == df.get_column("aika")
    assert roundtrip.get_column("alice").as_list() == df.get_column("alice")
    assert roundtrip.get_column("ai").as_list() == df.get_column("ai")


def test_data_frame_factor():
    df = BiocFrame({
        "regular": Factor.from_sequence([ "sydney", "melbourne", "", "perth", "adelaide" ]),
        "missing": Factor.from_sequence([ "sydney", None, "", "perth", "adelaide" ]),
        "ordered": Factor.from_sequence([ "sydney", "melbourne", "", "perth", "adelaide" ], ordered=True),
    })

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip["regular"], Factor)
    assert list(roundtrip["regular"]) == list(df["regular"])
    assert list(roundtrip["missing"]) == list(df["missing"])
    assert list(roundtrip["ordered"]) == list(df["ordered"])
    assert roundtrip["ordered"].get_ordered()


def test_data_frame_row_names():
    df = BiocFrame({
        "akari": [ 1, 2, 3, 4, 5 ],
        "aika": [ "sydney", "melbourne", "", "perth", "adelaide" ],
        "alice": [ True, False, False, True, True ],
        "ai": [ 2.3, 1.2, 5.2, 3.1, -1.2 ],
    }, row_names = [ "kaori", "chihaya", "fuyuki", "azusa", "iori" ])

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert df.row_names == roundtrip.row_names


def test_data_frame_wild_strings():
    df = BiocFrame({
        "lyrics": [ "Ochite\niku\tsunadokei\nbakari\tmiteru\tyo", "foo\"asdasd", "multie\"foo\"asdas" ],
    }, row_names = [ "nagisa\"", "\"fuko\"", "okazaki\nasdasd" ])

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert df.row_names == roundtrip.row_names
    assert df.get_column("lyrics") == roundtrip.get_column("lyrics").as_list()


def test_data_frame_none():
    df = BiocFrame({
        "akari": [ 1, 2, None, 4, 5 ],
        "aika": [ "sydney", None, "", "perth", "adelaide" ],
        "alice": [ True, None, False, None, True ],
        "ai": [ 2.3, None, 5.2, None, -1.2 ],
        "aria": [ None, None, None, None, None ],
        "akira": [ "south", None, "north", "east", "west" ],
    })

    def compare_masked_to_list(l, m):
        assert len(l) == len(m)
        for i, x in enumerate(l):
            if x is not None:
                assert x == m[i]
            else:
                assert np.ma.is_masked(m[i])

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    compare_masked_to_list(df.get_column("akari"), roundtrip.get_column("akari"))
    assert roundtrip.get_column("akari").dtype.type == np.int32

    assert roundtrip.get_column("aika").as_list() == df.get_column("aika")
    assert isinstance(roundtrip.get_column("aika"), StringList)

    compare_masked_to_list(df.get_column("alice"), roundtrip.get_column("alice"))
    assert roundtrip.get_column("alice").dtype.type == np.bool_

    compare_masked_to_list(df.get_column("ai"), roundtrip.get_column("ai"))
    assert roundtrip.get_column("ai").dtype.type == np.float64

    assert roundtrip.get_column("aria").as_list() == df.get_column("aria")

    assert roundtrip.get_column("akira").as_list() == df.get_column("akira")
    assert isinstance(roundtrip.get_column("akira"), StringList)


def test_data_frame_numpy():
    df = BiocFrame({
        "alicia": np.array([ 1, 2, 3, 4, 5 ]),
        "akira": np.array([ True, True, False, False, True ]),
        "athena": np.array([ 2.3, 2.3, 5.2, 32, -1.2 ]),
    })

    dir = os.path.join(mkdtemp(), "foo")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert (roundtrip.get_column("alicia") == df.get_column("alicia")).all()
    assert roundtrip.get_column("alicia").dtype == np.int32
    assert (roundtrip.get_column("akira") == df.get_column("akira")).all()
    assert roundtrip.get_column("akira").dtype == np.bool_
    assert (roundtrip.get_column("athena") == df.get_column("athena")).all()
    assert roundtrip.get_column("athena").dtype == np.float64

    # Coerces integers to floats.
    with h5py.File(os.path.join(dir, "basic_columns.h5"), "a") as handle:
        ghandle = handle["data_frame"]
        dhandle = ghandle["data"]
        xhandle = dhandle["0"]
        del xhandle.attrs["type"]
        xhandle.attrs.create("type", data="number")

    roundtrip = dl.read_object(dir)
    assert (roundtrip.get_column("alicia") == df.get_column("alicia")).all()
    assert roundtrip.get_column("alicia").dtype == np.float64


def test_data_frame_no_numpy():
    df = BiocFrame({
        "alicia": np.array([ 1, 2, 3, 4, 5 ]),
        "akira": np.array([ True, True, False, False, True ]),
        "athena": np.array([ 2.3, 2.3, 5.2, 32, -1.2 ]),
    })

    dir = os.path.join(mkdtemp(), "foo")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir, data_frame_represent_numeric_column_as_1darray=False)

    assert isinstance(roundtrip.get_column("alicia"), IntegerList)
    assert roundtrip.get_column("alicia").as_list() == list(df.get_column("alicia"))
    assert isinstance(roundtrip.get_column("akira"), BooleanList)
    assert roundtrip.get_column("akira").as_list() == list(df.get_column("akira"))
    assert isinstance(roundtrip.get_column("athena"), FloatList)
    assert roundtrip.get_column("athena").as_list() == list(df.get_column("athena"))


def test_data_frame_NamedList():
    df = BiocFrame({
        "alicia": IntegerList([ 1, 2, 3, 4, 5 ]),
        "akira": BooleanList([ True, True, False, False, True ]),
        "athena": FloatList([ 2.3, 2.3, 5.2, 32, -1.2 ]),
    })

    dir = os.path.join(mkdtemp(), "foo")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir, data_frame_represent_numeric_column_as_1darray=False)

    assert roundtrip.get_column("alicia") == df.get_column("alicia")
    assert roundtrip.get_column("akira") == df.get_column("akira")
    assert roundtrip.get_column("athena") == df.get_column("athena")


def test_data_frame_large_integers():
    df = BiocFrame({
        "alicia": [ 2**31 - 1, 1, 2, 3 ],
        "akira": [ 2**32, 1, 2, 3 ],
        "alice": np.array([ -2**31, 4, 5, 6 ], dtype=np.int64),
        "athena": np.array([ -2**32, 4, 5, 6 ], dtype=np.int64),
    })

    dir = os.path.join(mkdtemp(), "hdf5")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert (roundtrip.get_column("alicia") == df.get_column("alicia")).all()
    assert roundtrip.get_column("alicia").dtype == np.int32
    assert (roundtrip.get_column("akira") == df.get_column("akira")).all()
    assert roundtrip.get_column("akira").dtype == np.float64
    assert (roundtrip.get_column("alice") == df.get_column("alice")).all()
    assert roundtrip.get_column("alice").dtype == np.int32
    assert (roundtrip.get_column("athena") == df.get_column("athena")).all()
    assert roundtrip.get_column("athena").dtype == np.float64


def test_data_frame_special_floats():
    df = BiocFrame({
        "sumire": [ np.NaN, np.Inf, -np.Inf ],
        "kanon": np.array([ np.NaN, np.Inf, -np.Inf ]),
        "chisato": np.ma.array([ np.NaN, 4, 5 ], mask=[0, 1, 1]) # distinguish NaN from missing.
    })

    def as_expected(x):
        assert np.isnan(x[0])
        assert x[1] == np.Inf
        assert x[2] == -np.Inf

    def as_expected_masked(x):
        assert np.isnan(x[0])
        assert np.ma.is_masked(x[1])
        assert np.ma.is_masked(x[2])

    dir = os.path.join(mkdtemp(), "foo2")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, BiocFrame)
    as_expected(roundtrip.get_column("sumire"))
    as_expected(roundtrip.get_column("kanon"))
    as_expected_masked(roundtrip.get_column("chisato"))


def test_data_frame_masked():
    df = BiocFrame({
        "alicia": np.ma.array(np.array([ 1, 2, 3, 4, 5 ]), mask=[0, 1, 0, 1, 0]),
        "akira": np.ma.array(np.array([ True, True, False, False, True ]), mask=[1, 1, 0, 0, 0]), # important: test masking at the front.
        "athena": np.ma.array(np.array([ 2.3, -12.8, 5.2, 32, -1.2 ]), mask=[0, 0, 0, 1, 1]),
        "aika": np.ma.array(np.array([ 0, 0, 0, 0, 0 ]), mask=[1, 1, 1, 1, 1]),
    })

    dir = os.path.join(mkdtemp(), "foo")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, BiocFrame)
    assert (roundtrip.get_column("alicia") == df.get_column("alicia")).all()
    assert (roundtrip.get_column("alicia").mask == df.get_column("alicia").mask).all()
    assert (roundtrip.get_column("akira") == df.get_column("akira")).all()
    assert (roundtrip.get_column("akira").mask == df.get_column("akira").mask).all()
    assert (roundtrip.get_column("athena") == df.get_column("athena")).all()
    assert (roundtrip.get_column("athena").mask == df.get_column("athena").mask).all()
    assert (roundtrip.get_column("aika").mask == df.get_column("aika").mask).all()


def test_data_frame_empty():
    # Empty except for row names.
    df = BiocFrame({}, row_names=["chihaya", "mami", "ami", "miki", "haruka"])
    dir = os.path.join(mkdtemp(), "empty")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.row_names == roundtrip.row_names

    # Fully empty.
    df = BiocFrame(number_of_rows=10)
    dir = os.path.join(mkdtemp(), "empty")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BiocFrame)
    assert df.shape == roundtrip.shape
    assert roundtrip.row_names is None


def test_data_frame_nested():
    df = BiocFrame({
        "odd": [ 1, 3, 5, 7, 9 ],
        "liella": BiocFrame({ 
            "first": [ "kanon", "keke", "chisato", "sumire", "ren" ],
            "last": [ "shibuya", "tang", "arashi", "heanna", "hazuki" ],
            "best": [ False, False, False, True, False ],
        }),
        "even": [ 0, 2, 4, 6, 8 ],
        "bsb": BiocFrame({ 
            "first": [ "nick", "kevin", "brian", "AJ", "howie" ],
            "last": [ "carter", "richardson", "litrell", "maclean", "dorough" ]
        }),
    })

    dir = os.path.join(mkdtemp(), "foo2")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BiocFrame)

    liella_df = roundtrip.get_column("liella")
    assert isinstance(liella_df, BiocFrame)
    assert liella_df.get_column("first").as_list() == df.get_column("liella").get_column("first")
    assert liella_df.get_column("last").as_list() == df.get_column("liella").get_column("last")
    assert list(liella_df.get_column("best")) == df.get_column("liella").get_column("best")

    bsb_df = roundtrip.get_column("bsb")
    assert isinstance(bsb_df, BiocFrame)
    assert bsb_df.get_column("first").as_list() == df.get_column("bsb").get_column("first")
    assert bsb_df.get_column("last").as_list() == df.get_column("bsb").get_column("last")


def test_data_frame_metadata():
    df = BiocFrame(
        { "foo": [ 1, 3, 5, 7, 9 ] },
        column_data = BiocFrame({ "args": [ 99 ] }),
        metadata = { "a": 2, "b": ['a', 'b', 'c', 'd'] }
    )

    dir = os.path.join(mkdtemp(), "foo")
    dl.save_object(df, dir)
    roundtrip = dl.read_object(dir)
    roundtrip.metadata["b"] = roundtrip.metadata["b"].as_list()
    assert df.metadata == roundtrip.metadata
    assert roundtrip.get_column_data().get_column("args") == [ 99 ]
    assert isinstance(roundtrip, BiocFrame)

    # Any row names are deliberately stripped out.
    cd = df.get_column_data()
    cd2 = cd.set_row_names([ "ARGH" ])
    df2 = df.set_column_data(cd2)

    dir = os.path.join(mkdtemp(), "foo-empty")
    dl.save_object(df2, dir)
    roundtrip = dl.read_object(dir)
    assert roundtrip.get_column_data(with_names = False).get_row_names() is None



from biocframe import BiocFrame
import h5py
import os

from .alt_read_object import alt_read_object
from . import _utils_string as strings
from .load_vector_from_hdf5 import load_vector_from_hdf5
from ._utils_factor import load_factor_from_hdf5 
from . import _utils_misc as misc


def read_data_frame(path: str, metadata: dict, data_frame_represent_numeric_column_as_1darray : bool = True, **kwargs) -> BiocFrame:
    """Load a data frame from a HDF5 file. In general, this function should not
    be called directly but instead via :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path: 
            Path to the directory containing the object.

        metadata: 
            Metadata for the object.

        data_frame_represent_numeric_column_as_1darray: 
            Whether numeric columns should be represented as 1-dimensional
            NumPy arrays. This is more efficient than regular Python lists but
            discards the distinction between vectors and 1-D arrays. Usually
            this is not an important difference, but nonetheless, users can set
            this flag to ``False`` to load columns as (typed) lists instead.

        kwargs: Further arguments, passed to nested objects.

    Returns:
        A data frame.
    """
    column_names = []
    contents = {}
    row_names = None
    expected_rows = 0

    with h5py.File(os.path.join(path, "basic_columns.h5"), "r") as handle:
        ghandle = handle["data_frame"]
        expected_rows = ghandle.attrs["row-count"][()]
        column_names = strings.load_string_vector_from_hdf5(ghandle["column_names"])
        if "row_names" in ghandle:
            row_names = strings.load_string_vector_from_hdf5(ghandle["row_names"])

        dhandle = ghandle["data"]
        for i, col in enumerate(column_names):
            name = str(i)
            if name not in dhandle:
                contents[col] = alt_read_object(os.path.join(path, "other_columns", name), **kwargs)
            else:
                xhandle = dhandle[name]
                curtype = strings.load_scalar_string_attribute_from_hdf5(xhandle, "type")
                if curtype == "factor":
                    contents[col] = load_factor_from_hdf5(xhandle)
                else:
                    expected_type = misc.translate_type(curtype)
                    contents[col] = load_vector_from_hdf5(xhandle, expected_type, report_1darray=(expected_type != str and data_frame_represent_numeric_column_as_1darray))

    df = BiocFrame(contents, number_of_rows=expected_rows, row_names=row_names, column_names=column_names)

    other_dir = os.path.join(path, "other_annotations")
    if os.path.exists(other_dir):
        df.set_metadata(alt_read_object(other_dir, **kwargs).as_dict(), in_place=True)

    mcol_dir = os.path.join(path, "column_annotations")
    if os.path.exists(mcol_dir):
        df.set_column_data(alt_read_object(mcol_dir, **kwargs), in_place=True)

    return df



from typing import Sequence, Union
import numpy
import h5py
from biocutils import StringList, IntegerList, FloatList, BooleanList

from . import _utils_string as strings


def load_vector_from_hdf5(handle: h5py.Dataset, expected_type: type, report_1darray: bool) -> Union[StringList, IntegerList, FloatList, BooleanList, numpy.ndarray]:
    """
    Load a vector from a 1-dimensional HDF5 dataset, with coercion to the expected type.
    Any missing value placeholders are used to set Nones or to create masks.

    Args:
        handle: Handle to a HDF5 dataset.

        expected_type: 
            Expected type of the output vector. This should be one of
            ``float``, ``int``, ``str`` or ``bool``.

        report_1darray:
            Whether to report the output as a 1-dimensional NumPy array.

    Returns:
        The contents of the dataset as a vector-like object. By default, this
        is a typed :py:class:`~biocutils.biocutils.NamedList` subclass with
        missing values represented by None. If ``keep_as_1darray = True``, a
        1-dimensional NumPy array is returned instead, possibly with masking.
    """
    if expected_type == str:
        values = strings.load_string_vector_from_hdf5(handle)
        placeholder = None
        if "missing-value-placeholder" in handle.attrs:
            placeholder = strings.load_scalar_string_attribute_from_hdf5(handle, "missing-value-placeholder")
        if report_1darray:
            values = numpy.array(values)
            if placeholder is not None:
                mask = values == placeholder
                values = numpy.ma.MaskedArray(values, mask=mask)
        else:
            if placeholder is not None:
                for j, y in enumerate(values):
                    if y == placeholder:
                        values[j] = None
            values = StringList(values)
        return values

    values = handle[:]
    if "missing-value-placeholder" in handle.attrs:
        placeholder = handle.attrs["missing-value-placeholder"]
        if numpy.isnan(placeholder):
            mask = numpy.isnan(values)
        else:
            mask = (values == placeholder)

        if report_1darray:
            return numpy.ma.MaskedArray(_coerce_numpy_type(values, expected_type), mask=mask)
        else:
            output = []
            for i, y in enumerate(values):
                if mask[i]:
                    output.append(None)
                else:
                    output.append(y)
            return _choose_NamedList_subclass(output, expected_type)

    if report_1darray:
        return _coerce_numpy_type(values, expected_type)
    else:
        return _choose_NamedList_subclass(values, expected_type)


def _coerce_numpy_type(values: numpy.ndarray, expected_type: str) -> numpy.ndarray:
    if expected_type == bool:
        return values != 0
    elif expected_type == float:
        if not numpy.issubdtype(values.dtype, numpy.floating):
            return values.astype(numpy.double)
    return values


def _choose_NamedList_subclass(values: Sequence, expected_type: str) -> Union[IntegerList, FloatList, BooleanList]:
    if expected_type == bool:
        return BooleanList(values)
    elif expected_type == float:
        return FloatList(values)
    else:
        return IntegerList(values)



from biocutils import Factor
import os
import h5py

from .save_object import save_object, validate_saves
from . import _utils_string as strings
from ._utils_factor import save_factor_to_hdf5

@save_object.register
@validate_saves
def save_string_factor(x: Factor, path: str, **kwargs):
    """Method for saving :py:class:`~biocutils.Factor.Factor` objects to their
    corresponding file representation, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to save the object.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        handle.write('{ "type": "string_factor", "string_factor": { "version": "1.0" } }')

    with h5py.File(os.path.join(path, "contents.h5"), "w") as handle:
        ghandle = handle.create_group("string_factor")
        save_factor_to_hdf5(ghandle, x)
        nms = x.get_names()
        if nms is not None:
            strings.save_fixed_length_strings(ghandle, "names", nms.as_list())



from typing import Sequence, Union
import numpy
import h5py

from . import choose_missing_placeholder as ch
from . import _utils_misc as misc
from . import _utils_string as strings


def _is_missing_scalar(x): 
    return x is None or numpy.ma.is_masked(x)


def _has_missing(x: Sequence):
    if isinstance(x, numpy.ndarray):
        if isinstance(x, numpy.ma.MaskedArray):
            return x.mask.any()
        return False
    return any(_is_missing_scalar(y) for y in x)


def _fill_with_placeholder(x, dtype, placeholder):
    copy = numpy.ndarray(len(x), dtype=dtype)
    for i, y in enumerate(x):
        if _is_missing_scalar(y):
            copy[i] = placeholder
        else:
            copy[i] = y
    return copy


###################################################
###################################################


def write_string_vector_to_hdf5(
    handle: h5py.Group, 
    name: str, 
    x: Sequence[str], 
    placeholder_name: str = "missing-value-placeholder"
) -> h5py.Dataset:
    """
    Write a string vector to a HDF5 file as a 1-dimensional dataset with a
    fixed-length string datatype. If ``x`` contains missing values, a suitable
    placeholder value is selected using
    :py:func:`~dolomite_base.choose_missing_placeholder.choose_missing_string_placeholder`.
    and used to replace all missing values in the dataset. The placeholder
    itself is stored as an attribute of the dataset.

    Args:
        handle: A handle to a HDF5 group.

        name: Name of the dataset in which to save the string vector.

        x: Sequence containing strings, Nones, and/or masked NumPy values.
            
        placeholder_name: 
            Name of the attribute in which to store the missing value
            placeholder, if ``x`` contains None or masked values.

    Returns:
        Handle for the newly created dataset.
    """
    missed = _has_missing(x)

    if missed:
        placeholder = ch.choose_missing_string_placeholder(x)
        x = list(x)
        for i, y in enumerate(x):
            if _is_missing_scalar(y):
                x[i] = placeholder

    dset = strings.save_fixed_length_strings(handle, name, x)
    if missed:
        dset.attrs[placeholder_name] = placeholder
    return dset


###################################################
###################################################


def write_integer_vector_to_hdf5(
    handle: h5py.Group, 
    name: str, 
    x: Sequence[int], 
    h5type: str = "i4",
    placeholder_name: str = "missing-value-placeholder", 
    allow_float_promotion: bool = False 
) -> h5py.Dataset:
    """
    Write an integer vector to a HDF5 file as a 1-dimensional dataset. If
    ``x`` contains missing values, a placeholder value is selected by
    :py:func:`~dolomite_base.choose_missing_placeholder.choose_missing_integer_placeholder`
    and used to replace all of the missing values in the dataset. The
    placeholder value itself is stored as an attribute of the dataset.

    Args:
        handle: A handle to a HDF5 group.

        name: Name of the dataset in which to save the integer vector.

        x: Sequence containing integers, Nones, and/or masked NumPy values.

        h5type: Integer type of the HDF5 dataset to create.
            
        placeholder_name: 
            Name of the attribute in which to store the missing value
            placeholder, if ``x`` contains None or masked values.

        allow_float_promotion:
            Whether to save ``x`` into a 64-bit floating-point dataset if any
            values in ``x`` exceeds the range of values that can be represented
            by ``h5type``, or if no missing value placeholder can be found
            within the acceptable range of integer values. If ``False``, an
            error is raised if ``x`` cannot be saved without promotion.

    Returns:
        Handle for the newly created dataset.
    """
    missed = _has_missing(x)

    max_dtype = numpy.dtype(h5type).type
    limits = numpy.iinfo(max_dtype)
    exceeds = False
    for y in x:
        if not _is_missing_scalar(y):
            if y < limits.min or y > limits.max:
                exceeds = True
                break

    if exceeds:
        if not allow_float_promotion:
            raise ValueError("cannot save out-of-range integers without type promotion")
        if missed:
            placeholder = numpy.NaN
            x = _fill_with_placeholder(x, numpy.float64, placeholder)
    else:
        if missed:
            placeholder = ch.choose_missing_integer_placeholder(x, max_dtype=max_dtype)
            if placeholder is None:
                exceeds = True
                if not allow_float_promotion:
                    raise ValueError("cannot find a suitable missing value placeholder without type promotion")
                placeholder = numpy.NaN
                x = _fill_with_placeholder(x, numpy.float64, placeholder)
            else:
                x = _fill_with_placeholder(x, placeholder.dtype.type, placeholder)

    if exceeds:
        h5type = "f8"

    dset = handle.create_dataset(name, data=x, dtype=h5type, compression="gzip", chunks=True)
    if missed:
       dset.attrs.create(placeholder_name, placeholder, dtype=h5type)
    return dset


###################################################
###################################################


def write_float_vector_to_hdf5(
    handle: h5py.Group, 
    name: str, 
    x: Sequence[float], 
    h5type: str = "f8",
    placeholder_name: str = "missing-value-placeholder"
) -> h5py.Dataset:
    """
    Write a floating-point vector to a HDF5 file as a 1-dimensional dataset.
    If ``x`` contains missing values, a placeholder value is selected by
    :py:func:`~dolomite_base.choose_missing_placeholder.choose_missing_float_placeholder`.
    and used to replace all of the missing values in the dataset. The
    placeholder value itself is stored as an attribute of the dataset.

    Args:
        handle: A handle to a HDF5 group.

        name: Name of the dataset in which to save the integer vector.

        x: Sequence containing floats, Nones, and/or masked NumPy values.

        h5type: Floating-point type of the HDF5 dataset to create.
            
        placeholder_name: 
            Name of the attribute in which to store the missing value
            placeholder, if ``x`` contains None or masked values.

    Returns:
        Handle for the newly created dataset.
    """
    missed = _has_missing(x)
    if missed:
        dtype = numpy.dtype(h5type).type
        placeholder = ch.choose_missing_float_placeholder(x, dtype=dtype)
        x = _fill_with_placeholder(x, dtype, placeholder)

    dset = handle.create_dataset(name, data=x, dtype=h5type, compression="gzip", chunks=True)
    if missed:
       dset.attrs.create(placeholder_name, placeholder, dtype=h5type)
    return dset


###################################################
###################################################


def write_boolean_vector_to_hdf5(
    handle: h5py.Group, 
    name: str, 
    x: Sequence[bool],
    placeholder_name: str = "missing-value-placeholder"
) -> h5py.Dataset:
    """
    Write a boolean vector to a HDF5 file as a 1-dimensional dataset with
    a 8-bit signed integer datatype. If ``x`` contains missing values, they
    are replaced with a placeholder value of -1.

    Args:
        handle: A handle to a HDF5 group.

        name: Name of the dataset in which to save the integer vector.

        x: Sequence containing booleans, Nones, and/or masked NumPy values.
            
        placeholder_name: 
            Name of the attribute in which to store the missing value
            placeholder, if ``x`` contains None or masked values.

    Returns:
        Handle for the newly created dataset.
    """

    missed = _has_missing(x)
    if missed:
        placeholder = -1
        x = _fill_with_placeholder(x, numpy.int8, placeholder)

    h5type = "i1"
    dset = handle.create_dataset(name, data=x, dtype=h5type, compression="gzip", chunks=True)
    if missed:
       dset.attrs.create(placeholder_name, placeholder, dtype=h5type)
    return dset




from typing import Any, Optional
from importlib import import_module
import os
import json

registry = {
    "atomic_vector": "dolomite_base.read_atomic_vector",
    "string_factor": "dolomite_base.read_string_factor",
    "simple_list": "dolomite_base.read_simple_list",
    "data_frame": "dolomite_base.read_data_frame"
}


def read_object(path: str, metadata: Optional[dict] = None, **kwargs) -> Any:
    """Read an object from its on-disk representation.

    Args:
        path: 
            Path to a directory containing the object.

        metadata: 
            Metadata for the object. This is read from the `OBJECT` file if None.

        kwargs: 
            Further arguments, passed to individual methods.

    Returns:
        Some kind of object.
    """
    if metadata is None:
        with open(os.path.join(path, "OBJECT"), "rb") as handle:
            metadata = json.load(handle)

    tt = metadata["type"]
    if tt not in registry:
        raise NotImplementedError("could not find a Python command to read '" + tt + "'")

    command = registry[tt]
    if isinstance(command, str): 
        first_period = command.find(".")
        mod = import_module(command[:first_period])
        command = getattr(mod, command[first_period + 1:])
        registry[tt] = command

    return command(path, metadata, **kwargs)



from typing import Any
from functools import singledispatch, wraps
from .validate_object import validate_object


@singledispatch
def save_object(x: Any, path: str, **kwargs):
    """Save an object to its on-disk representation. 
    
    **dolomite** extensions should define methods for 
    this generic to stage different object classes.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to the output directory.

        kwargs: 
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    raise NotImplementedError("'save_object' is not implemented for " + str(type(x)))


def validate_saves(fn):
    """
    Decorator to validate the output of :py:func:`~save_object`.

    Args:
        fn: Function that implements a method for ``save_object``.

    Returns:
        A wrapped version of the function that validates the directory
        containing the on-disk representation of the saved object.
    """
    @wraps(fn)
    def wrapper(x, path, **kwargs):
        out = fn(x, path, **kwargs)
        validate_object(path)
        return out
    return wrapper



from biocutils import Factor
import h5py
import os

from ._utils_factor import load_factor_from_hdf5
from . import _utils_string as strings


def read_string_factor(path: str, metadata: dict, **kwargs) -> Factor:
    """Read a string factor from disk. 
    
    In general, this function should not be called directly 
    but instead via
    :py:meth:`~dolomite_base.load_object.load_object`.

    Args:
        path: 
            Path to the directory containing the object.

        metadata: 
            Metadata for the object. 

        kwargs: 
            Further arguments, passed to nested objects.

    Returns:
        A :py:class:`~biocutils.Factor.Factor` object.
    """

    with h5py.File(os.path.join(path, "contents.h5"), "r") as handle:
        ghandle = handle["string_factor"]
        output = load_factor_from_hdf5(ghandle)
        if "names" in ghandle:
            output.set_names(strings.load_string_vector_from_hdf5(ghandle["names"]), in_place=True)
        return output



import numpy
from biocutils import Factor
import h5py

from . import _utils_string as strings


def save_factor_to_hdf5(handle: h5py.Group, f: Factor):
    strings.save_fixed_length_strings(handle, "levels", f.get_levels())

    codes = f.get_codes()
    is_missing = codes == -1
    has_missing = is_missing.any()
    nlevels = len(f.get_levels())
    if has_missing:
        codes = codes.astype(numpy.uint32, copy=True)
        codes[is_missing] = nlevels

    dhandle = handle.create_dataset("codes", data=codes, dtype="u4", compression="gzip", chunks=True)
    if has_missing:
        dhandle.attrs.create("missing-value-placeholder", data=nlevels, dtype="u4")

    if f.get_ordered():
        handle.attrs.create("ordered", data=1, dtype="i1")


def load_factor_from_hdf5(handle: h5py.Group):
    chandle = handle["codes"]
    codes = chandle[:]
    codes = codes.astype(numpy.int32, copy=False)

    if "missing-value-placeholder" in chandle.attrs:
        placeholder = chandle.attrs["missing-value-placeholder"]
        codes[codes == placeholder] = -1

    ordered = False
    if "ordered" in handle.attrs:
        ordered = handle.attrs["ordered"][()] != 0
    
    levels = strings.load_string_vector_from_hdf5(handle["levels"])
    return Factor(codes, levels, ordered = ordered)



import dolomite_base as dl
from biocutils import StringList, IntegerList, FloatList, BooleanList
from tempfile import mkdtemp
import h5py
import numpy
import os
import pytest

def test_string_list_basic():
    sl = StringList([1,2,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, StringList)
    assert roundtrip == sl

    sl = StringList([1,2,None,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, StringList)
    assert roundtrip == sl


def test_string_list_names():
    sl = StringList([1,2,3,4])
    sl = sl.set_names(["A", "B", "C", "D"])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, StringList)
    assert roundtrip == sl


def test_integer_list_basic():
    sl = IntegerList([1,2,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, IntegerList)
    assert roundtrip == sl

    sl = IntegerList([1,2,None,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, IntegerList)
    assert roundtrip == sl


def test_integer_list_ndarray():
    sl = IntegerList([1,2,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ndarray)
    assert roundtrip2.dtype == numpy.int32
    assert list(roundtrip2) == sl.as_list()

    sl = IntegerList([1,2,None,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ma.MaskedArray)
    assert roundtrip2.dtype == numpy.int32
    assert [None if numpy.ma.is_masked(y) else int(y) for y in roundtrip2] == sl.as_list()


def test_integer_list_large():
    sl = IntegerList([2**32,4**20])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, FloatList)
    assert roundtrip == sl


def test_integer_list_names():
    sl = IntegerList([1,2,3,4])
    sl = sl.set_names(["A", "B", "C", "D"])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, IntegerList)
    assert roundtrip == sl

    with pytest.warns(UserWarning, match="skipping names"):
        roundtrip = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)


def test_float_list_basic():
    sl = FloatList([1,2,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, FloatList)
    assert roundtrip == sl

    sl = FloatList([1,2,None,3,4])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, FloatList)
    assert roundtrip == sl


def test_float_list_ndarray():
    sl = FloatList([1.2,2.3,3.4,4.5])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ndarray)
    assert roundtrip2.dtype == numpy.float64
    assert list(roundtrip2) == sl.as_list()

    sl = FloatList([1.2,None,2.3,None,3.4,4.5])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ma.MaskedArray)
    assert roundtrip2.dtype == numpy.float64
    assert [None if numpy.ma.is_masked(y) else float(y) for y in roundtrip2] == sl.as_list()


def test_float_list_names():
    sl = FloatList([1,2,3,4])
    sl = sl.set_names(["A", "B", "C", "D"])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, FloatList)
    assert roundtrip == sl


def test_boolean_list_basic():
    sl = BooleanList([True, False, False, True])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BooleanList)
    assert roundtrip == sl

    sl = BooleanList([True, False, None, False, True])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BooleanList)
    assert roundtrip == sl


def test_boolean_list_ndarray():
    sl = BooleanList([True, False, False, True])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ndarray)
    assert roundtrip2.dtype == numpy.bool_
    assert list(roundtrip2) == sl.as_list()

    sl = BooleanList([True, False, None, False, True])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip2 = dl.read_object(dir, atomic_vector_use_numeric_1darray=True)
    assert isinstance(roundtrip2, numpy.ma.MaskedArray)
    assert roundtrip2.dtype == numpy.bool_
    assert [None if numpy.ma.is_masked(y) else bool(y) for y in roundtrip2] == sl.as_list()


def test_boolean_list_names():
    sl = BooleanList([True, False, False, True])
    sl = sl.set_names(["A", "B", "C", "D"])
    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(sl, dir)
    roundtrip = dl.read_object(dir)
    assert isinstance(roundtrip, BooleanList)
    assert roundtrip == sl



from typing import Any, Callable, Dict, Optional

from .save_object import save_object

ALT_SAVE_OBJECT_FUNCTION = save_object


def alt_save_object_function(fun: Optional[Callable] = None) -> Optional[Callable]:
    """Get or set the alternative saving function for use by
    :py:meth:`~alt_save_object`. Typically set by applications prior to
    saving for customization, e.g., to save extra metadata.

    Args:
        fun:
            The alternative saving function. This should accept the same
            arguments and return the same value as
            :py:meth:`~dolomite_base.save_object.save_object`.

    Returns:
        If ``fun = None``, the current setting of the alternative saving
        function is returned. 

        Otherwise, the alternative saving function is set to ``fun``,
        and the previous function is returned.
    """
    global ALT_SAVE_OBJECT_FUNCTION
    if fun is None:
        return ALT_SAVE_OBJECT_FUNCTION
    else:
        old = ALT_SAVE_OBJECT_FUNCTION
        ALT_SAVE_OBJECT_FUNCTION = fun
        return old


def alt_save_object(x: Any, path: str, **kwargs) -> Dict[str, Any]:
    """Wrapper around :py:meth:`~dolomite_base.save_object.save_object` that
    respects application-defined overrides from
    :py:meth:`~alt_save_object_function`.  
    
    This allows applications to
    customize the saving process for some or all of the object classes,
    assuming that developers of dolomite extensions (and the associated
    ``save_object`` methods) use ``alt_save_object`` internally for saving
    child objects instead of ``save_object``.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to a directory to save `x`.

        kwargs: 
            Further arguments to be passed to individual methods.

    Returns:
        `x` is saved to `path`.
    """
    fun = alt_save_object_function()
    return fun(x, path, **kwargs)



import dolomite_base as dl
from tempfile import mkdtemp
import h5py
import numpy
import os
import pytest


def test_write_string_vector_to_hdf5():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_string_vector_to_hdf5(ghandle, "FOO1", ["A", "B", "C", "D"])
        dl.write_string_vector_to_hdf5(ghandle, "FOO2", ["A", "B", None, "D"])
        dl.write_string_vector_to_hdf5(ghandle, "FOO3", numpy.array(["A", "B", "C", "D"]))
        dl.write_string_vector_to_hdf5(ghandle, "FOO4", numpy.ma.MaskedArray(numpy.array(["A", "B", "NA", "D"]), mask=[True, False, False, False]))

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]
        assert [x.decode() for x in ghandle["FOO1"]] == ["A", "B", "C", "D"]
        assert [x.decode() for x in ghandle["FOO2"]] == ["A", "B", "NA", "D"]
        assert ghandle["FOO2"].attrs["missing-value-placeholder"] == "NA"
        assert [x.decode() for x in ghandle["FOO3"]] == ["A", "B", "C", "D"]
        assert [x.decode() for x in ghandle["FOO4"]] == ["NA_", "B", "NA", "D"]
        assert ghandle["FOO4"].attrs["missing-value-placeholder"] == "NA_"


def test_write_integer_vector_to_hdf5_simple():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_integer_vector_to_hdf5(ghandle, "FOO1", [1, 2, 3, 4])
        dl.write_integer_vector_to_hdf5(ghandle, "FOO2", [1, 2, None, 3])
        dl.write_integer_vector_to_hdf5(ghandle, "FOO3", numpy.array([1, 2, 3, 4], dtype=numpy.uint8))
        dl.write_integer_vector_to_hdf5(ghandle, "FOO4", numpy.ma.MaskedArray(numpy.array([1, 2, 3, 4], dtype=numpy.uint8), mask=[True, False, False, False]))

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]
        assert ghandle["FOO1"].dtype == "i4"
        assert list(ghandle["FOO1"]) == [1, 2, 3, 4]
        assert list(ghandle["FOO2"]) == [1, 2, -2**31, 3]
        assert ghandle["FOO2"].attrs["missing-value-placeholder"] == -2**31
        assert ghandle["FOO3"].dtype == "i4"
        assert list(ghandle["FOO3"]) == [1, 2, 3, 4]
        assert list(ghandle["FOO4"]) == [0, 2, 3, 4]
        assert ghandle["FOO4"].attrs["missing-value-placeholder"] == 0 


def test_write_integer_vector_to_hdf5_exceeds_limit():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")

        with pytest.raises(Exception, match="out-of-range"):
            dl.write_integer_vector_to_hdf5(ghandle, "FOO1", [1, 2, -2**32, 4])
        dl.write_integer_vector_to_hdf5(ghandle, "FOO1", [1, 2, -2**32, 4], allow_float_promotion=True)

        y = numpy.ma.MaskedArray(numpy.array(range(257), dtype=numpy.uint32), mask=[False]*256 + [True])
        with pytest.raises(Exception, match="suitable missing value placeholder"):
            dl.write_integer_vector_to_hdf5(ghandle, "FOO2", y, h5type="u1")
        dl.write_integer_vector_to_hdf5(ghandle, "FOO2", y, h5type="u1", allow_float_promotion=True)

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]
        assert ghandle["FOO1"].dtype == "f8"
        assert list(ghandle["FOO1"]) == [1, 2, -2**32, 4]

        assert ghandle["FOO2"].dtype == "f8"
        assert numpy.isnan(ghandle["FOO2"][-1])
        assert list(ghandle["FOO2"][:-1]) == list(range(256))
        assert numpy.isnan(ghandle["FOO2"].attrs["missing-value-placeholder"])


def test_write_float_vector_to_hdf5_simple():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_float_vector_to_hdf5(ghandle, "FOO1", [1.5, 2.2, 3.5, 4.7])
        dl.write_float_vector_to_hdf5(ghandle, "FOO2", [1.5, 2.2, 4.7, None])
        dl.write_float_vector_to_hdf5(ghandle, "FOO3", numpy.array([1, 3, 2, 4], dtype=numpy.float32))
        dl.write_float_vector_to_hdf5(ghandle, "FOO4", numpy.ma.MaskedArray(numpy.array([1, 2, 3, 4], dtype=numpy.float32), mask=[True, False, False, False]))

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]
        assert ghandle["FOO1"].dtype == "f8"
        assert list(ghandle["FOO1"]) == [1.5, 2.2, 3.5, 4.7]

        assert list(ghandle["FOO2"][:-1]) == [1.5, 2.2, 4.7]
        assert numpy.isnan(ghandle["FOO2"][-1])
        assert numpy.isnan(ghandle["FOO2"].attrs["missing-value-placeholder"])

        assert ghandle["FOO3"].dtype == "f8"
        assert list(ghandle["FOO3"]) == [1, 3, 2, 4]

        assert list(ghandle["FOO4"][1:]) == [2, 3, 4]
        assert numpy.isnan(ghandle["FOO4"][0])
        assert numpy.isnan(ghandle["FOO4"].attrs["missing-value-placeholder"])


def test_write_boolean_vector_to_hdf5_simple():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO1", [True, False, False, True])
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO2", [False, False, True, None])
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO3", numpy.array([True, False, True, False], dtype=numpy.bool_))
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO4", numpy.ma.MaskedArray(numpy.array([False, True, True, False], dtype=numpy.bool_), mask=[True, False, False, False]))

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]
        assert ghandle["FOO1"].dtype == "i1"
        assert list(ghandle["FOO1"]) == [1, 0, 0, 1]
        assert list(ghandle["FOO2"]) == [0, 0, 1, -1]
        assert ghandle["FOO2"].attrs["missing-value-placeholder"] == -1
        assert list(ghandle["FOO3"]) == [1, 0, 1, 0]
        assert ghandle["FOO4"].dtype == "i1"
        assert list(ghandle["FOO4"]) == [-1, 1, 1, 0]
        assert ghandle["FOO4"].attrs["missing-value-placeholder"] == -1



from . import lib_dolomite_base as lib


def validate_object(path: str):
    """Validate an on-disk representation of an object,
    using validators based on the **takane** specifications.

    Args:
        path: 
            Path to the directory containing the object's representation.

    Raise:
        Error if the validation fails.
    """
    lib.validate(path)
    return



import dolomite_base as dl
from biocutils import StringList, Factor
from tempfile import mkdtemp
import os
import numpy


def test_string_factor():
    regular = Factor.from_sequence([ "sydney", "melbourne", "brisbane", "perth", "adelaide" ])

    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(regular, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, Factor)
    assert (roundtrip.get_codes() == regular.get_codes()).all()
    assert roundtrip.get_levels() == regular.get_levels()
    assert roundtrip.get_ordered() == regular.get_ordered()


def test_string_factor_missing():
    missing = Factor.from_sequence([ "sydney", None, None, None, None ])

    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(missing, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, Factor)
    assert (roundtrip.get_codes() == missing.get_codes()).all()
    assert roundtrip.get_levels() == missing.get_levels()
    assert roundtrip.get_ordered() == missing.get_ordered()


def test_string_factor_ordered():
    ordered = Factor.from_sequence([ "sydney", "melbourne", "brisbane", "perth", "adelaide" ], ordered=True, levels=["sydney", "melbourne", "adelaide", "perth", "brisbane"])

    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(ordered, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, Factor)
    assert (roundtrip.get_codes() == ordered.get_codes()).all()
    assert roundtrip.get_levels() == ordered.get_levels()
    assert roundtrip.get_ordered() == ordered.get_ordered()


def test_string_factor_named():
    named = Factor.from_sequence([ "sydney", "melbourne", "brisbane", "perth", "adelaide" ])
    named.set_names(["A", "B", "C", "D", "E"], in_place=True)

    dir = os.path.join(mkdtemp(), "temp")
    dl.save_object(named, dir)
    roundtrip = dl.read_object(dir)

    assert isinstance(roundtrip, Factor)
    assert roundtrip.get_names() == named.get_names()



from typing import Any, Optional, Callable
from .read_object import read_object


ALT_READ_OBJECT_FUNCTION = read_object


def alt_read_object_function(fun: Optional[Callable] = None) -> Optional[Callable]:
    """
    Get or set the alternative reading function for use by
    :py:meth:`~alt_read_object`. Typically set by applications prior to reading
    for customization, e.g., to attach more metadata to the readed object.

    Args:
        fun:
            The alternative reading function. This should accept the same
            arguments and return the same value as
            :py:meth:`~dolomite_base.read_object.read_object`.

    Returns:
        If ``fun = None``, the current setting of the alternative reading
        function is returned. 

        Otherwise, the alternative reading function is set to ``fun``,
        and the previous function is returned.
    """
    global ALT_READ_OBJECT_FUNCTION
    if fun is None:
        return ALT_READ_OBJECT_FUNCTION
    else:
        old = ALT_READ_OBJECT_FUNCTION
        ALT_READ_OBJECT_FUNCTION = fun
        return old


def alt_read_object(path: str, **kwargs) -> Any:
    """Wrapper around :py:meth:`~dolomite_base.read_object.read_object` 
    that respects application-defined overrides from
    :py:meth:`~alt_read_object_function`.  This allows applications to
    customize the reading process for some or all of the object classes,
    assuming that developers of dolomite extensions (and the associated
    functions called by ``read_object``) use ``alt_read_object`` internally for
    staging child objects instead of ``read_object``.

    Args:
        path: 
            Directory containing the object to load.

        kwargs: 
            Further arguments, passed to individual methods.

    Returns:
        Some kind of object.
    """
    fun = alt_read_object_function()
    return fun(path, **kwargs)



LIMIT32 = 2**31


def scalar_exceeds_int32(x: int) -> bool:
    return x < -LIMIT32 or x >= LIMIT32


def sequence_exceeds_int32(x: int, check_none: bool = True) -> bool:
    if check_none:
        for y in x:
            if y is not None and scalar_exceeds_int32(y):
                return True
    else:
        for y in x:
            if scalar_exceeds_int32(y):
                return True
    return False


def translate_type(t: str) -> type:
    if t == "string":
        return str
    elif t == "number":
        return float 
    elif t == "integer":
        return int
    elif t == "boolean":
        return bool
    else:
        raise NotImplementedError("unknown vector type '" + t + "'")



"""
    Setup file for dolomite_base.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 4.5.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig
from glob import glob
import pathlib
import os
import shutil
import sys
import pybind11

## Adapted from https://stackoverflow.com/questions/42585210/extending-setuptools-extension-to-use-cmake-in-setup-py.
class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])

class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)

    def build_cmake(self, ext):
        build_temp = pathlib.Path(self.build_temp)
        build_lib = pathlib.Path(self.build_lib)
        outpath = os.path.join(build_lib.absolute(), ext.name) 

        if not os.path.exists(build_temp):
            cmd = [ 
                "cmake", 
                "-S", "lib",
                "-B", build_temp,
                "-Dpybind11_DIR=" + os.path.join(os.path.dirname(pybind11.__file__), "share", "cmake", "pybind11"),
                "-DPYTHON_EXECUTABLE=" + sys.executable
            ]
            if os.name != "nt":
                cmd.append("-DCMAKE_BUILD_TYPE=Release")
                cmd.append("-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + outpath)

            if "MORE_CMAKE_OPTIONS" in os.environ:
                cmd += os.environ["MORE_CMAKE_OPTIONS"].split()
            self.spawn(cmd)

        if not self.dry_run:
            cmd = ['cmake', '--build', build_temp]
            if os.name == "nt":
                cmd += ["--config", "Release"]
            self.spawn(cmd)
            if os.name == "nt": 
                # Gave up trying to get MSVC to respect the output directory.
                # Delvewheel also needs it to have a 'pyd' suffix... whatever.
                shutil.copyfile(os.path.join(build_temp, "Release", "_core.dll"), os.path.join(outpath, "_core.pyd"))

if __name__ == "__main__":
    import os
    try:
        setup(
            use_scm_version={"version_scheme": "no-guess-dev"},
            ext_modules=[CMakeExtension("dolomite_base")],
            cmdclass={
                'build_ext': build_ext
            }
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "dolomite-base"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .save_object import save_object, validate_saves
from .validate_object import validate_object
from .save_atomic_vector import save_atomic_vector_from_string_list, save_atomic_vector_from_integer_list, save_atomic_vector_from_float_list, save_atomic_vector_from_boolean_list
from .save_string_factor import save_string_factor
from .save_simple_list import save_simple_list_from_list, save_simple_list_from_dict, save_simple_list_from_NamedList
from .save_data_frame import save_data_frame
from .read_atomic_vector import read_atomic_vector
from .read_string_factor import read_string_factor
from .read_simple_list import read_simple_list
from .read_data_frame import read_data_frame
from .read_object import read_object

from .choose_missing_placeholder import choose_missing_integer_placeholder, choose_missing_float_placeholder, choose_missing_string_placeholder
from .write_vector_to_hdf5 import write_string_vector_to_hdf5, write_float_vector_to_hdf5, write_integer_vector_to_hdf5, write_boolean_vector_to_hdf5
from .load_vector_from_hdf5 import load_vector_from_hdf5



from biocutils import StringList, IntegerList, FloatList, BooleanList
import os
import h5py
import numpy

from .save_object import save_object, validate_saves
from . import _utils_string as strings
from . import write_vector_to_hdf5 as write


@save_object.register
@validate_saves
def save_atomic_vector_from_string_list(x: StringList, path: str, **kwargs): 
    """Method for saving :py:class:`~biocutils.StringList.StringList` objects
    to their corresponding file representation, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to save the object.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        handle.write('{ "type": "atomic_vector", "atomic_vector": { "version": "1.0" } }')

    with h5py.File(os.path.join(path, "contents.h5"), "w") as handle:
        ghandle = handle.create_group("atomic_vector")
        ghandle.attrs["type"] = "string"
        write.write_string_vector_to_hdf5(ghandle, "values", x.as_list())
        nms = x.get_names()
        if nms is not None:
            strings.save_fixed_length_strings(ghandle, "names", nms.as_list())

    return


@save_object.register
@validate_saves
def save_atomic_vector_from_integer_list(x: IntegerList, path: str, **kwargs): 
    """Method for saving :py:class:`~biocutils.IntegerList.IntegerList` objects
    to their corresponding file representation, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to save the object.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        handle.write('{ "type": "atomic_vector", "atomic_vector": { "version": "1.0" } }')

    with h5py.File(os.path.join(path, "contents.h5"), "w") as handle:
        ghandle = handle.create_group("atomic_vector")
        dset = write.write_integer_vector_to_hdf5(ghandle, "values", x.as_list(), allow_float_promotion=True)

        if numpy.issubdtype(dset, numpy.floating):
            ghandle.attrs["type"] = "number"
            dset.attrs.create("_python_original_type", "biocutils.IntegerList")
        else:
            ghandle.attrs["type"] = "integer"

        nms = x.get_names()
        if nms is not None:
            strings.save_fixed_length_strings(ghandle, "names", nms.as_list())

    return


@save_object.register
@validate_saves
def save_atomic_vector_from_float_list(x: FloatList, path: str, **kwargs): 
    """Method for saving :py:class:`~biocutils.FloatList.FloatList` objects
    to their corresponding file representation, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to save the object.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        handle.write('{ "type": "atomic_vector", "atomic_vector": { "version": "1.0" } }')

    with h5py.File(os.path.join(path, "contents.h5"), "w") as handle:
        ghandle = handle.create_group("atomic_vector")
        ghandle.attrs["type"] = "number"
        write.write_float_vector_to_hdf5(ghandle, "values", x.as_list())
        nms = x.get_names()
        if nms is not None:
            strings.save_fixed_length_strings(ghandle, "names", nms.as_list())

    return


@save_object.register
@validate_saves
def save_atomic_vector_from_boolean_list(x: BooleanList, path: str, **kwargs): 
    """Method for saving :py:class:`~biocutils.BooleanList.BooleanList` objects
    to their corresponding file representation, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to save the object.

        kwargs: 
            Further arguments, ignored.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)
    with open(os.path.join(path, "OBJECT"), 'w', encoding="utf-8") as handle:
        handle.write('{ "type": "atomic_vector", "atomic_vector": { "version": "1.0" } }')

    with h5py.File(os.path.join(path, "contents.h5"), "w") as handle:
        ghandle = handle.create_group("atomic_vector")
        ghandle.attrs["type"] = "boolean"
        write.write_boolean_vector_to_hdf5(ghandle, "values", x.as_list())
        nms = x.get_names()
        if nms is not None:
            strings.save_fixed_length_strings(ghandle, "names", nms.as_list())

    return



import dolomite_base as dl
from tempfile import mkdtemp
import h5py
import numpy
import os
import biocutils


def test_load_vector_from_hdf5_strings():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_string_vector_to_hdf5(ghandle, "FOO1", ["A", "B", "C", "D"])
        dl.write_string_vector_to_hdf5(ghandle, "FOO2", ["A", "B", None, "D"])

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]

        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], str, report_1darray=False)
        assert isinstance(foo1, biocutils.StringList)
        assert foo1.as_list() == ["A", "B", "C", "D"]
        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], str, report_1darray=True)
        assert isinstance(foo1, numpy.ndarray)
        assert list(foo1) == ["A", "B", "C", "D"]

        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], str, report_1darray=False)
        assert isinstance(foo2, biocutils.StringList)
        assert foo2.as_list() == ["A", "B", None, "D"]
        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], str, report_1darray=True)
        assert isinstance(foo2, numpy.ndarray)
        assert list(foo2.data) == ["A", "B", "NA", "D"]
        assert list(foo2.mask) == [False, False, True, False]


def test_load_vector_from_hdf5_integers():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_integer_vector_to_hdf5(ghandle, "FOO1", [1, 2, 3, 4])
        dl.write_integer_vector_to_hdf5(ghandle, "FOO2", [1, 2, None, 4])
        dl.write_integer_vector_to_hdf5(ghandle, "FOO3", [1, 2, 3, 2**32], h5type="u8")

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]

        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], int, report_1darray=False)
        assert isinstance(foo1, biocutils.IntegerList)
        assert foo1.as_list() == [1, 2, 3, 4]
        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], int, report_1darray=True)
        assert isinstance(foo1, numpy.ndarray)
        assert foo1.dtype == numpy.int32
        assert list(foo1) == [1, 2, 3, 4]

        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], int, report_1darray=False)
        assert isinstance(foo2, biocutils.IntegerList)
        assert foo2.as_list() == [1, 2, None, 4]
        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], int, report_1darray=True)
        assert isinstance(foo2, numpy.ndarray)
        assert foo2.dtype == numpy.int32
        assert list(foo2.data) == [1, 2, -2**31, 4]
        assert list(foo2.mask) == [False, False, True, False]

        foo3 = dl.load_vector_from_hdf5(ghandle["FOO3"], int, report_1darray=False)
        assert isinstance(foo3, biocutils.IntegerList)
        assert foo3.as_list() == [1, 2, 3, 2**32]
        foo3 = dl.load_vector_from_hdf5(ghandle["FOO3"], int, report_1darray=True)
        assert isinstance(foo3, numpy.ndarray)
        assert foo3.dtype == numpy.uint64
        assert list(foo3) == [1, 2, 3, 2**32]


def test_load_vector_from_hdf5_floats():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_float_vector_to_hdf5(ghandle, "FOO1", [1.1, 2.2, 3.3, 4.4])
        dl.write_float_vector_to_hdf5(ghandle, "FOO2", [1.1, 2.2, 4.4, None])
        dl.write_float_vector_to_hdf5(ghandle, "FOO3", [1.1, 2.2, 4.4, None, numpy.NaN]) # check for correct behavior with non-NaN placeholders
        dl.write_integer_vector_to_hdf5(ghandle, "FOO4", [1, 2, 3, 2**32], h5type="u8") # check for correct promotion of integer storage types.

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]

        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], float, report_1darray=False)
        assert isinstance(foo1, biocutils.FloatList)
        assert foo1.as_list() == [1.1, 2.2, 3.3, 4.4]
        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], float, report_1darray=True)
        assert isinstance(foo1, numpy.ndarray)
        assert list(foo1) == [1.1, 2.2, 3.3, 4.4]

        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], float, report_1darray=False)
        assert isinstance(foo2, biocutils.FloatList)
        assert foo2.as_list() == [1.1, 2.2, 4.4, None]
        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], float, report_1darray=True)
        assert isinstance(foo2, numpy.ndarray)
        assert list(foo2.data[:-1]) == [1.1, 2.2, 4.4]
        assert numpy.isnan(foo2.data[-1])
        assert list(foo2.mask) == [False, False, False, True]

        foo3 = dl.load_vector_from_hdf5(ghandle["FOO3"], float, report_1darray=False)
        assert isinstance(foo3, biocutils.FloatList)
        assert foo3.as_list()[:-1] == [1.1, 2.2, 4.4, None]
        assert numpy.isnan(foo3.as_list()[-1])
        foo3 = dl.load_vector_from_hdf5(ghandle["FOO3"], float, report_1darray=True)
        assert isinstance(foo3, numpy.ndarray)
        assert list(foo3.data[:-1]) == [1.1, 2.2, 4.4, numpy.inf]
        assert numpy.isnan(foo3.data[-1])
        assert list(foo3.mask) == [False, False, False, True, False]

        foo4 = dl.load_vector_from_hdf5(ghandle["FOO4"], float, report_1darray=False)
        assert isinstance(foo4, biocutils.FloatList)
        assert foo4.as_list() == [1, 2, 3, 2**32]
        foo4 = dl.load_vector_from_hdf5(ghandle["FOO4"], float, report_1darray=True)
        assert isinstance(foo4, numpy.ndarray)
        assert list(foo4) == [1, 2, 3, 2**32]


def test_load_vector_from_hdf5_booleans():
    path = os.path.join(mkdtemp(), "foo.h5")
    with h5py.File(path, "w") as handle:
        ghandle = handle.create_group("yourmom")
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO1", [True, False, False, True])
        dl.write_boolean_vector_to_hdf5(ghandle, "FOO2", [True, True, None, False])

    with h5py.File(path, "r") as handle:
        ghandle = handle["yourmom"]

        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], bool, report_1darray=False)
        assert isinstance(foo1, biocutils.BooleanList)
        assert foo1.as_list() == [True, False, False, True]
        foo1 = dl.load_vector_from_hdf5(ghandle["FOO1"], bool, report_1darray=True)
        assert isinstance(foo1, numpy.ndarray)
        assert foo1.dtype == numpy.bool_
        assert list(foo1) == [True, False, False, True]

        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], bool, report_1darray=False)
        assert isinstance(foo2, biocutils.BooleanList)
        assert foo2.as_list() == [True, True, None, False]
        foo2 = dl.load_vector_from_hdf5(ghandle["FOO2"], bool, report_1darray=True)
        assert isinstance(foo2, numpy.ndarray)
        assert list(foo2.data) == [True, True, True, False]
        assert list(foo2.mask) == [False, False, True, False]



from typing import Any, Dict
from collections import namedtuple
from functools import singledispatch
import os
from biocframe import BiocFrame
from biocutils import Factor, StringList, IntegerList, BooleanList, FloatList
import numpy
import h5py

from .save_object import save_object
from .alt_save_object import alt_save_object
from . import _utils_string as strings
from . import write_vector_to_hdf5 as write
from ._utils_factor import save_factor_to_hdf5


@save_object.register
def save_data_frame(
    x: BiocFrame, 
    path: str, 
    data_frame_convert_list_to_vector: bool = True, 
    data_frame_convert_1darray_to_vector: bool = True, 
    **kwargs
) -> Dict[str, Any]:
    """Method for saving :py:class:`~biocframe.BiocFrame.BiocFrame`
    objects to the corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x: 
            Object to be saved.

        path: 
            Path to a directory in which to save ``x``.

        data_frame_convert_list_to_vector: 
            If a column is a regular Python list where all entries are of the
            same basic type (integer, string, float, boolean) or None, should
            it be converted to a typed vector in the on-disk representation?
            This avoids creating a separate file to store this column but
            changes the class of the column when the ``BiocFrame`` is read back
            into a Python session. If ``False``, the list is saved as an
            external object instead.

        data_frame_convert_1darray_to_vector: 
            If a column is a 1D NumPy array, should it be saved as a typed
            vector? This avoids creating a separate file for the column but
            discards the distinction between 1D arrays and vectors. Usually
            this is not an important difference, but nonetheless, users can
            set this flag to ``False`` to save all 1D NumPy arrays as an 
            external "dense array" object instead.

        kwargs: 
            Further arguments, passed to internal
            :py:func:`~dolomite_base.alt_save_object.alt_save_object` calls.

    Returns:
        `x` is saved to `path`.
    """
    os.mkdir(path)

    other = []
    full = os.path.join(path, "basic_columns.h5")
    with h5py.File(full, "w") as handle:
        ghandle = handle.create_group("data_frame")
        ghandle.attrs.create("row-count", data=x.shape[0], dtype="u8")
        ghandle.attrs.create("version", data="1.0")

        dhandle = ghandle.create_group("data")
        output = Hdf5ColumnOutput(
            handle=dhandle, 
            otherable=other, 
            convert_list_to_vector=data_frame_convert_list_to_vector, 
            convert_1darray_to_vector=data_frame_convert_1darray_to_vector
        )
        for i in range(x.shape[1]):
            _process_column_for_hdf5(x.get_column(i), i, output)

        strings.save_fixed_length_strings(ghandle, "column_names", x.get_column_names())
        rn = x.get_row_names()
        if rn is not None:
            strings.save_fixed_length_strings(ghandle, "row_names", rn)

    if len(other):
        other_dir = os.path.join(path, "other_columns")
        os.mkdir(other_dir)
        for i in other:
            alt_save_object(x.get_column(i), os.path.join(other_dir, str(i)), data_frame_convert_list_to_vector=data_frame_convert_list_to_vector, **kwargs)

    md = x.get_metadata()
    if md is not None and len(md):
        alt_save_object(x.metadata, os.path.join(path, "other_annotations"), data_frame_convert_list_to_vector=data_frame_convert_list_to_vector, **kwargs)

    cd = x.get_column_data(with_names=False) 
    if cd is not None and cd.shape[1] > 0:
        if cd.get_row_names() is not None:
            cd = cd.set_row_names(None)
        alt_save_object(cd, os.path.join(path, "column_annotations"), data_frame_convert_list_to_vector=data_frame_convert_list_to_vector, **kwargs)

    with open(os.path.join(path, "OBJECT"), "w") as handle:
        handle.write('{ "type": "data_frame", "data_frame": { "version": "1.0" } }')
    return


########################################################


Hdf5ColumnOutput = namedtuple('Hdf5ColumnOutput', ['handle', 'otherable', 'convert_list_to_vector', 'convert_1darray_to_vector'])


@singledispatch
def _process_column_for_hdf5(x: Any, index: int, output: Hdf5ColumnOutput):
    output.otherable.append(index)
    return


@_process_column_for_hdf5.register
def _process_list_column_for_hdf5(x: list, index: int, output: Hdf5ColumnOutput):
    if output.convert_list_to_vector:
        all_types = set()
        has_none = False
        for y in x:
            if isinstance(y, numpy.generic):
                if numpy.issubdtype(y.dtype.type, numpy.integer):
                    all_types.add(int)
                elif numpy.issubdtype(y.dtype.type, numpy.floating):
                    all_types.add(float)
                else:
                    all_types.add(type(y))
            elif y is None:
                has_none = True
            else:
                all_types.add(type(y))

        final_type = None
        if len(all_types) == 1:
            final_type = list(all_types)[0]
        elif len(all_types) == 2 and int in all_types and float in all_types:
            final_type = float
        elif len(all_types) == 0:
            final_type = None

        if final_type == str:
            dhandle = write.write_string_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "string"
            return

        elif final_type == int:
            dhandle = write.write_integer_vector_to_hdf5(output.handle, str(index), x, allow_float_promotion=True)
            if numpy.issubdtype(dhandle.dtype, numpy.floating):
                dhandle.attrs["type"] = "number"
            else:
                dhandle.attrs["type"] = "integer"
            return

        elif final_type == float:
            dhandle = write.write_float_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "number"
            return

        elif final_type == bool:
            dhandle = write.write_boolean_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "boolean"
            return

    _process_column_for_hdf5.registry[object](x, index, output)
    return


@_process_column_for_hdf5.register
def _process_StringList_column_for_hdf5(x: StringList, index: int, output: Hdf5ColumnOutput):
    dhandle = write.write_string_vector_to_hdf5(output.handle, str(index), x.as_list())
    dhandle.attrs["type"] = "string"
    return


@_process_column_for_hdf5.register
def _process_IntegerList_column_for_hdf5(x: IntegerList, index: int, output: Hdf5ColumnOutput):
    dhandle = write.write_integer_vector_to_hdf5(output.handle, str(index), x.as_list(), allow_float_promotion=True)
    if numpy.issubdtype(dhandle.dtype, numpy.floating):
        dhandle.attrs["type"] = "number"
    else:
        dhandle.attrs["type"] = "integer"
    return


@_process_column_for_hdf5.register
def _process_FloatList_column_for_hdf5(x: FloatList, index: int, output: Hdf5ColumnOutput):
    dhandle = write.write_float_vector_to_hdf5(output.handle, str(index), x.as_list())
    dhandle.attrs["type"] = "number"
    return


@_process_column_for_hdf5.register
def _process_BooleanList_column_for_hdf5(x: BooleanList, index: int, output: Hdf5ColumnOutput):
    dhandle = write.write_float_vector_to_hdf5(output.handle, str(index), x.as_list())
    dhandle.attrs["type"] = "boolean"
    return


@_process_column_for_hdf5.register
def _process_ndarray_column_for_hdf5(x: numpy.ndarray, index: int, output: Hdf5ColumnOutput):
    if output.convert_1darray_to_vector and len(x.shape) == 1:
        if numpy.issubdtype(x.dtype, numpy.floating):
            dhandle = write.write_float_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "number"
            return

        elif x.dtype == numpy.bool_:
            dhandle = write.write_boolean_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "boolean"
            return

        elif numpy.issubdtype(x.dtype, numpy.integer):
            dhandle = write.write_integer_vector_to_hdf5(output.handle, str(index), x, allow_float_promotion=True)
            if numpy.issubdtype(dhandle.dtype, numpy.floating):
                dhandle.attrs["type"] = "number"
            else:
                dhandle.attrs["type"] = "integer"
            return

        elif numpy.issubdtype(x.dtype, numpy.str_):
            dhandle = write.write_string_vector_to_hdf5(output.handle, str(index), x)
            dhandle.attrs["type"] = "string"

        else:
            raise NotImplementedError("cannot save column of type '" + x.dtype.name + "'")

    _process_column_for_hdf5.registry[object](x, index, output)
    return


@_process_column_for_hdf5.register
def _process_factor_column_for_hdf5(x: Factor, index: int, output: Hdf5ColumnOutput):
    ghandle = output.handle.create_group(str(index))
    ghandle.attrs.create("type", data="factor")
    save_factor_to_hdf5(ghandle, x)
    return
